# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

import sixtracktools
import xtrack as xt
import xobjects as xo
import xfields as xf

# sixtrack_folder = './sixtrack_lhc_no_bb/res_onmom';
# atol = {
#     'x':1e-16, 'px':1e-16, 'y':1e-16, 'py':1e-16, 'zeta':6e-14, 'delta':1e-15}

#sixtrack_folder = './sixtrack_lhc_no_bb/res_offmom';
#atol = {
#    'x':1e-16, 'px':1e-16, 'y':1e-16, 'py':1e-16, 'zeta':5e-14, 'delta':1e-15}

sixtrack_folder = '../../test_data/hllhc_14/res'; atol = 5e-12
atol = {
    'x':1e-15, 'px':2e-12, 'y':1e-15, 'py':2e-12, 'zeta':2e-14, 'delta':1e-15}

context = xo.ContextCpu()

sixinput = sixtracktools.SixInput(sixtrack_folder)
line = sixinput.generate_xtrack_line()
line.build_tracker(_context=context)
iconv = line.other_info["iconv"]

sixdump_all = sixtracktools.SixDump101(sixtrack_folder + "/dump3.dat")

if any(ee.__class__.__name__.startswith('BeamBeam') for ee in line.elements):
    Nele_st = len(iconv)
    sixdump_CO = sixdump_all[::2][:Nele_st]
    # Get closed-orbit from sixtrack 
    p0c_eV = sixinput.initialconditions[-3] * 1e6
    part_on_CO = xt.Particles(
            p0c=p0c_eV,
            x=sixdump_CO.x[0],
            px=sixdump_CO.px[0],
            y=sixdump_CO.y[0],
            py=sixdump_CO.py[0],
            zeta=sixdump_CO.tau[0]*sixdump_CO.beta0[0],
            delta=sixdump_CO.delta[0])
    xf.configure_orbit_dependent_parameters_for_bb(line,
                           particle_on_co=part_on_CO)

def compare(prun, pbench):
    out = []
    for att in "x px y py zeta delta".split():
        vrun = getattr(prun, att)[0]
        vbench = getattr(pbench, att)[0]
        diff = vrun - vbench
        out.append(abs(diff))
        print(f"{att:<5} {vrun:22.13e} {vbench:22.13e} {diff:22.13g}")
    print(f"max {max(out):21.12e}")
    return max(out), out

sixdump = sixdump_all[1::2]
print("")
diffs = []
s_coord = []
x_run = []
x_bench = []
s_run = []
s_bench = []
for ii in range(1, len(iconv)):
    jja = iconv[ii - 1]
    jjb = iconv[ii]
    prun = xt.Particles.from_dict(_context=context,
            dct=sixdump[ii - 1].get_minimal_beam())
    prun.state[0]=1
    prun.reorganize()
    print(f"\n-----sixtrack={ii} xtrack={jja} --------------")
    # print(f"pysixtr {jja}, x={prun.x}, px={prun.px}")
    for jj in range(jja + 1, jjb + 1):
        label, elem = line.element_names[jj], line.elements[jj]
        #elem.track(prun)
        line.track(particles=prun, ele_start=jj, num_elements=1)
        print(f"{jj} {label},{str(elem)[:50]}")
    pbench = xt.Particles.from_dict(sixdump[ii].get_minimal_beam())
    s_coord.append(pbench.s)
    x_run.append(prun.x)
    x_bench.append(pbench.x)
    s_run.append(prun.s)
    s_bench.append(pbench.s)
    # print(f"sixdump {ii}, x={pbench.x}, px={pbench.px}")
    print("-----------------------")
    out, out_all = compare(prun, pbench)
    print("-----------------------\n\n")
    diffs.append(out_all)
    issue_found = False
    for ii, nn in enumerate('x px y py zeta delta'.split()):
        if out_all[ii] > atol[nn]:
            print("Too large discrepancy")
            issue_found = True

    if issue_found:
        break

diffs = np.array(diffs)

import matplotlib.pyplot as plt
plt.close('all')
fig = plt.figure(1, figsize=(6.4*1.5, 4.8*1.3))
for ii, (vv, uu) in enumerate(
        zip(['x', 'px', 'y', 'py', r'$\zeta$', r'$\delta$'],
            ['[m]', '[-]', '[m]', '[-]', '[m]', '[-]'])):
    ax = fig.add_subplot(3, 2, ii+1)
    ax.plot(s_coord, diffs[:, ii])
    ax.set_ylabel('Difference on '+ vv + ' ' + uu)
    ax.set_xlabel('s [m]')
fig.subplots_adjust(hspace=.48)


plt.show()



import xtrack as xt

env = xt.load_madx_lattice('../../test_data/ps_sftpro/ps.seq')
env.vars.load_madx('../../test_data/ps_sftpro/ps_hs_sftpro.str')

line = env.ps
line.particle_ref = xt.Particles(kinetic_energy0=14e9)

line.twiss4d().plot()


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np
import xtrack as xt
from scipy.constants import c as clight

line = xt.Line.from_json('../../test_data/sps_w_spacecharge/line_no_spacecharge.json')
line.build_tracker()
tw0 = line.twiss4d()

f0 = 1 / (line.get_length()/(clight*line.particle_ref.beta0[0]))
h_rf = 4620

# Remove existing voltage from all cavities
tt = line.get_table(attr=True)
tt_cav = tt.rows[tt.element_type=='Cavity']
for nn in tt_cav.name:
    line[nn].voltage = 0

# Define knobs controlling the RF cavities
line['on_dfreq'] = 1.
line['on_rf1'] = 1.
line['on_rf2'] = 1.
line['dfreq'] = 1.e3 * line.ref['on_dfreq'] # Frequency shift in Hz
line['v_rf1'] = 1.2e6 * line.ref['on_rf1']
line['v_rf2'] = 1.2e6 * line.ref['on_rf2']

# Setup two RF cavities with different frequencies
line['acta.31637'].absolute_time = 1 # <-- define cavity w.r.t. absolute time and not to ref. particle!!!
line['acta.31637'].frequency = h_rf * f0 + line.ref['dfreq']
line['acta.31637'].voltage = line.ref['v_rf1']
line['acta.31637'].lag = 180

line['actd.31934'].absolute_time = 1 # <-- define cavity w.r.t. absolute time and not to ref. particle!!!
line['actd.31934'].frequency = h_rf * f0 - line.ref['dfreq']
line['actd.31934'].voltage = line.ref['v_rf2']
line['actd.31934'].lag = 180

# Display the cavity properties
tt = line.get_table(attr=True)
tt_cav = tt.rows[['acta.31637', 'actd.31934']]
tt_cav.cols['frequency voltage lag'].show(digits=10)

# Twiss (looking for actual revolution frequency)
line['on_rf2'] = 0.
tw_rf1 = line.twiss(search_for_t_rev=True)
line['on_rf2'] = 1.

line['on_rf1'] = 0.
tw_rf2 = line.twiss(search_for_t_rev=True)
line['on_rf1'] = 1.

print('Revolution frequency with RF1 only: ', 1/tw_rf1.T_rev)
print('Revolution frequency with RF2 only: ', 1/tw_rf2.T_rev)
print('Energy deviation with RF1 only: ', tw_rf1.ptau[0])
print('Energy deviation with RF2 only: ', tw_rf2.ptau[0])

# match a train (no frequency shift and single RF)
dfreq_tmp = line['dfreq']
line['dfreq'] = 0
line['on_rf2'] = 0.
import xpart as xp
train0 = xp.generate_matched_gaussian_multibunch_beam(
    sigma_z=0.1, nemitt_x=0, nemitt_y=0, line=line,
    bunch_intensity_particles=1e11,
    bunch_num_particles=500, filling_scheme=5*[1, 0, 0, 0, 0])
line['on_rf2'] = 1.
line['dfreq'] = dfreq_tmp

# Make two trains matched to the two RF systems
train1 = train0.copy()
train1.delta += tw_rf1.delta[0]
train2 = train0.copy()
train2.delta += tw_rf2.delta[0]
bucket_length = line.get_length() / h_rf
train2.zeta += 20 * bucket_length

# Make a single particle set with the two trains
two_trains = xt.Particles.merge([train1, train2])

# Track!
num_turns = 400
line.track(two_trains, num_turns=num_turns, with_progress=10,
           turn_by_turn_monitor=True)
mon = line.record_last_track

# Plots
import matplotlib.pyplot as plt
plt.close('all')

# Plot one turn
i_turn = 350
plt.figure(4)
plt.plot(mon.zeta.T[i_turn, :], mon.delta.T[i_turn, :], '.')
plt.show()

# Make movie (needed `conda install -c conda-forge ffmpeg``)
def update_plot(i_turn, fig):
    plt.clf()
    plt.plot(mon.zeta.T[i_turn, :], mon.delta.T[i_turn, :], '.', markersize=1)
    plt.xlim(-40, 40)
    plt.ylim(-12e-3, 12e-3)
    plt.xlabel('z [m]')
    plt.ylabel(r'$\Delta p / p_0$')
    plt.title(f'Turn {i_turn}')
    plt.subplots_adjust(left=0.2)
    plt.grid(alpha=0.5)

fig = plt.figure()
from matplotlib.animation import FFMpegFileWriter
moviewriter = FFMpegFileWriter(fps=15)
with moviewriter.saving(fig, 'slipstack.mp4', dpi=100):
    for j in range(0, num_turns, 2):
        update_plot(j, fig)
        moviewriter.grab_frame()



import xtrack as xt
import numpy as np

theta_tilt_deg = 20
theta_tilt_rad = theta_tilt_deg * np.pi / 180

l_solenoid = 3

l_beam = l_solenoid / np.cos(theta_tilt_rad)

tilt_entry = xt.YRotation(angle=-theta_tilt_deg)
tilt_exit = xt.YRotation(angle=+theta_tilt_deg)
shift_entry = xt.XYShift(dx=l_solenoid/2 * np.tan(theta_tilt_rad))
shift_exit = xt.XYShift(dx=l_solenoid/2 * np.tan(theta_tilt_rad))
solenoid = xt.Drift(length=l_solenoid)

line = xt.Line(
    elements=[xt.Marker(), tilt_entry, shift_entry, solenoid, shift_exit, tilt_exit, xt.Marker()],
    element_names=['start', 'tilt_entry', 'shift_entry', 'solenoid', 'shift_exit', 'tilt_exit', 'end']
)
line.config.XTRACK_USE_EXACT_DRIFTS = True
line.config.XTRACK_GLOBAL_XY_LIMIT = 1000.

line2 = xt.Line(
    elements=[xt.Marker(), xt.Drift(length=l_beam), xt.Marker()],
    element_names=['start', 'solenoid', 'end']
)
line2.config.XTRACK_USE_EXACT_DRIFTS = True
line2.config.XTRACK_GLOBAL_XY_LIMIT = 1000.
line2.build_tracker()



line.build_tracker()
line.tracker.skip_end_turn_actions = True


# particle on the solenoid axis
p0 = xt.Particles(p0c=6500e9, px=-np.sin(theta_tilt_rad),
                  x=l_beam/2 * np.tan(theta_tilt_rad),
                  delta=0.1, py=0.1)
p = p0.copy()
line.track(p)

p2 = p0.copy()
line2.track(p2)


pref = xt.Particles(p0c=6500e9)
line.particle_ref = pref
line2.particle_ref = pref
tw = line.twiss(betx=1, bety=1, x=p0.x[0], px=p0.px[0], start='start', end='end')
tw2 = line2.twiss(betx=1, bety=1, x=p0.x[0], px=p0.px[0], start='start', end='end')

pco = pref.copy()
R = line.compute_one_turn_matrix_finite_differences(particle_on_co=pco)['R_matrix']



import xobjects as xo
import xtrack as xt
from xobjects.test_helpers import for_all_test_contexts

@for_all_test_contexts
def test_amplitude_detuning(test_context):
    line = xt.Line(elements=[xt.LineSegmentMap(qx=62.31, qy=60.32,
                            det_xx=1000, det_xy=10, det_yx=20, det_yy=2000)])
    line.particle_ref = xt.Particles(p0c=7e9)
    line.build_tracker(_context=test_context)

    det = line.get_amplitude_detuning_coefficients()

    xo.assert_allclose(det['det_xx'], 1000, atol=1e-1, rtol=0)
    xo.assert_allclose(det['det_yy'], 2000, atol=1e-1, rtol=0)
    xo.assert_allclose(det['det_xy'], 10, atol=1e-1, rtol=0)
    xo.assert_allclose(det['det_yx'], 20, atol=1e-1, rtol=0)



import pandas as pd
import numpy as np

from scipy.constants import c as clight

import xtrack as xt
import xdeps as xd

# REMEMBER:
# - Handle zero ramp rate

fname = '../../test_data/psb_chicane/Ramp_and_RF_functions.dat'

df = pd.read_csv(fname, sep='\t', skiprows=2,
    names=['t_s', 'E_kin_GeV', 'V1_MV', 'phi1_rad', 'V2_MV', 'phi2_rad'])
E_kin_GeV = df.E_kin_GeV.values
t_s = df.t_s.values

# # Stretch it to enhance change in revolution frequency
# E_min = np.min(E_kin_GeV)
# E_max = np.max(E_kin_GeV)
# E_kin_GeV = E_min/100 + (E_kin_GeV - E_min)
# # Shift the time scale for testing purposes
# t_s = t_s + 5e-3

line = xt.Line.from_json('psb_04_with_chicane_corrected_thin.json')
line.build_tracker()
line['br1.acwf7l1.1'].voltage = 2e3

# Attach energy program
ep = xt.EnergyProgram(t_s=t_s, kinetic_energy0=E_kin_GeV*1e9)
line.energy_program = ep

f_rev = line.energy_program.get_frev_at_t_s(t_s)

line.functions['fun_f_rev'] = xd.FunctionPieceWiseLinear(x=t_s, y=f_rev)
line.vars['f_rev'] = line.functions['fun_f_rev'](line.vars['t_turn_s'])
line.element_refs['br1.acwf7l1.1'].frequency = line.vars['f_rev'] # cavity on h=1

tw6d = line.twiss(method='6d')

t_rev = []
qs = []
zeta_co = []
beta0 = []
for ii in range(len(t_s)):
    print(f'Computing twiss at t_s = {t_s[ii]:.4} s    ', end='\r', flush=True)
    line.vars['t_turn_s'] = t_s[ii]
    tt = line.twiss(method='6d')
    t_rev.append(tt.T_rev0)
    qs.append(tt.qs)
    zeta_co.append(tt.zeta[0])
    beta0 = tt.beta0

line.vars['t_turn_s'] = 0

t_rev = np.array(t_rev)
qs = np.array(qs)
zeta_co = np.array(zeta_co)
beta0 = np.array(beta0)

f1 = 1/t_rev
phis_1_deg = 2 * np.pi * f1 * zeta_co / beta0 / clight * 360 / 2 / np.pi

line.functions['fun_phi1'] = xd.FunctionPieceWiseLinear(x=t_s, y=-phis_1_deg)
line.vars['phi1_deg'] = line.functions['fun_phi1'](line.vars['t_turn_s'])

line.element_refs['br1.acwf7l1.1'].lag = line.vars['phi1_deg']

tw = line.twiss()

# Test tracking
p_test = line.build_particles(x_norm=0, zeta=np.linspace(0, 100., 20))
assert np.isclose(p_test.energy0[0] - p_test.mass0,  E_kin_GeV[0] * 1e9,
                  atol=0, rtol=1e-10)

line.enable_time_dependent_vars = True
n_turn_test = 10000
monitor = xt.ParticlesMonitor(num_particles=len(p_test.zeta), start_at_turn=0,
                              stop_at_turn=n_turn_test)
for ii in range(n_turn_test):
    if ii % 10 == 0:
        print(f'Tracking turn {ii}/{n_turn_test}     ', end='\r', flush=True)
    line.track(p_test, turn_by_turn_monitor=monitor)


t_test = 40e-3
p0c_test = ep.get_p0c_at_t_s(t_test)
p_test.update_p0c_and_energy_deviations(p0c_test)
ekin_test = p_test.energy0[0] - p_test.mass0

etot_expected_ev = E_kin_GeV * 1e9 + p_test.mass0
gamma_expected = etot_expected_ev / p_test.mass0
beta_expected = np.sqrt(1 - 1/gamma_expected**2)

t_turn = line.energy_program.get_t_s_at_turn(np.arange(n_turn_test))

import matplotlib.pyplot as plt
plt.close('all')

plt.figure(1)
sp_ekin = plt.subplot(3,1,1)
plt.plot(t_s, E_kin_GeV)
plt.plot(t_test, ekin_test*1e-9, 'o')
plt.ylabel(r'$E_{kin}$ [GeV]')

sp_dekin = plt.subplot(3,1,2, sharex=sp_ekin)
# GeV/sec
dekin = (E_kin_GeV[1:] - E_kin_GeV[:-1])/(t_s[1:] - t_s[:-1])
plt.plot(t_s[:-1], dekin)
plt.ylabel(r'd$E_{kin}$/dt [GeV/s]')

sp_beta = plt.subplot(3,1,3, sharex=sp_ekin)
plt.plot(t_s, beta_expected, '--', color='k', alpha=0.4)
plt.plot(t_turn, monitor.beta0.T)
plt.ylabel(r'$\beta$')
plt.xlabel('t [s]')

plt.figure(2)
i_turn_test = np.arange(n_turn_test)
t_turn_constant_energy = i_turn_test * line.get_length()/clight/beta_expected[0]
plt.plot(i_turn_test, 1e6 * (t_turn - t_turn_constant_energy))

i_turn_test_method = 9000
t_test_set = ep.get_t_s_at_turn(i_turn_test_method)
t_turn_method_constant_energy = i_turn_test_method * line.get_length()/clight/beta_expected[0]
plt.plot(i_turn_test_method, 1e6 * (t_test_set - t_turn_method_constant_energy), 'o')

plt.ylabel(r't - L / ($\beta_{inj} c$) [$\mu$s]')

plt.figure(3)
colors = plt.cm.jet(np.linspace(0,1, len(p_test.zeta)))
for ii in range(len(colors)):
    plt.plot(monitor.zeta[ii, :], monitor.delta[ii, :], color=colors[ii])
plt.xlim(-300, 300)
plt.ylim(-5e-3, 5e-3)

plt.show()



import xtrack as xt
import numpy as np

line = xt.Line.from_json('../../test_data/hllhc15_thick/lhc_thick_with_knobs.json')
line.twiss_default.update({'strengths': False, 'method': '4d'})
tw0 = line.twiss()


tt = line.get_table()

observable_list = ['betx', 'bety', 'mux', 'muy', 'dx']

obs_points = tt.rows['bpm.*'].name
corr_names = line.vars.get_table().rows['kq.*.b1'].name

response = {oo: np.zeros((len(obs_points), len(corr_names)))
            for oo in observable_list}

dk = 0.5e-5
for ii, nn in enumerate(corr_names):
    print(f'Processing {ii}/{len(corr_names)}')
    line.vars[nn] += dk
    twp = line.twiss()
    for jj, mm in enumerate(obs_points):
        for observable in observable_list:
            response[observable][jj, ii] = (twp[observable, mm] - tw0[observable, mm]) / dk
    line.vars[nn] -= dk


line.vars['kq7.r5b1'] *= 1.01
tw = line.twiss()

tw['betx0'] = tw0['betx']
tw['bety0'] = tw0['bety']
tw['mux0'] = tw0['mux']
tw['muy0'] = tw0['muy']

from xtrack.trajectory_correction import _compute_correction

corr_on_observable = 'mux'

err = tw.rows[obs_points][corr_on_observable] - tw0.rows[obs_points][corr_on_observable]
response_matrix = response[corr_on_observable]

correction_svd = _compute_correction(err, response_matrix, rcond=1e-2)
correction_micado = _compute_correction(err, response_matrix, n_micado=1)

i_micado = np.argmax(np.abs(correction_micado))
print(f'MICADO correction: {correction_micado[i_micado]:.2e} at {corr_names[i_micado]}')

import matplotlib.pyplot as plt
plt.close('all')
plt.figure(1)
plt.plot(correction_svd, '.', label='SVD')
plt.plot(correction_micado, '.', label='MICADO')

plt.legend()
plt.show()


import numpy as np
import xtrack as xt
import xobjects as xo
from cpymad.madx import Madx

test_context = xo.context_default

mad = Madx()

dr_len = 1e-11
mad.input(f"""
ss: sequence, l={dr_len};
    lens: nllens, at=0, cnll=0.15, knll=0.3;
    ! since in MAD-X we can't track a zero-length line, we put in
    ! this tiny drift here at the end of the sequence:
    dr: drift, at={dr_len / 2}, l={dr_len};
endsequence;
beam;
use, sequence=ss;
""")

line = xt.Line.from_madx_sequence(mad.sequence.ss)
line.config.XTRACK_USE_EXACT_DRIFTS = True # to be consistent with madx
line.build_tracker(_context=test_context)

num_p_test = 10
x_test = np.linspace(-1e-2, 2e-2, num_p_test)
y_test = np.linspace(-3e-2, 1e-2, num_p_test)
px_test = np.linspace(-2e-5, 4e-5, num_p_test)
py_test = np.linspace(-4e-5, 2e-5, num_p_test)

p0 = xt.Particles(p0c=2e9, x=x_test, px=px_test, y=y_test, py=py_test,
                  zeta=.1, ptau=1e-3)

part = p0.copy(_context=test_context)
line.track(part, _force_no_end_turn_actions=True)
part.move(_context=xo.context_default)

xt_tau = part.zeta/part.beta0
px = []
py = []
for ii in range(len(p0.x)):
    mad.input(f"""
    beam, particle=proton, pc={p0.p0c[ii] / 1e9}, sequence=ss, radiate=FALSE;

    track, onepass, onetable;
    start, x={p0.x[ii]}, px={p0.px[ii]}, y={p0.y[ii]}, py={p0.py[ii]}, \
        t={p0.zeta[ii]/p0.beta0[ii]}, pt={p0.ptau[ii]};
    run,
        turns=1,
        track_harmon=1e-15;  ! since in this test we don't care about
            ! losing particles due to t difference, we set track_harmon to
            ! something very small, to make t_max large.
    endtrack;
    """)

    mad_results = mad.table.mytracksumm[-1]

    px.append(mad_results.px)
    py.append(mad_results.py)


    assert np.allclose(part.x[ii], mad_results.x, atol=1e-14, rtol=0), 'x'
    assert np.allclose(part.px[ii], mad_results.px, atol=1e-14, rtol=0), 'px'
    assert np.allclose(part.y[ii], mad_results.y, atol=1e-14, rtol=0), 'y'
    assert np.allclose(part.py[ii], mad_results.py, atol=1e-14, rtol=0), 'py'
    assert np.allclose(xt_tau[ii], mad_results.t, atol=1e-14, rtol=0), 't'
    assert np.allclose(part.ptau[ii], mad_results.pt, atol=1e-14, rtol=0), 'pt'
    assert np.allclose(part.s[ii], mad_results.s, atol=1e-14, rtol=0), 's'

import matplotlib.pyplot as plt
plt.close('all')

plt.figure(1)
plt.plot(x_test, px, '.')
plt.plot(x_test, part.px, 'x')
plt.plot(x_test, py, '.')
plt.plot(x_test, part.py, 'x')
plt.show()




from .particles_monitor import *
from .last_turns_monitor import *
from .beam_position_monitor import *
from .beam_size_monitor import *
from .beam_profile_monitor import *

monitor_classes = tuple(v for v in globals().values() if isinstance(v, type) and issubclass(v, BeamElement))



import numpy as np

import xtrack as xt


fringe = xt.DipoleEdge(k=0.12, fint=100, hgap=0.035, model='full')

line = xt.Line(elements=[fringe])
line.particle_ref = xt.Particles(mass0=xt.PROTON_MASS_EV, beta0=0.5)
line.build_tracker()

p0 = line.build_particles(px=0.5, py=0.001, y=0.01, delta=0.1)

p_ng = p0.copy()
p_ptc = p0.copy()

R_ng = line.compute_one_turn_matrix_finite_differences(particle_on_co=p0.copy())
line.track(p_ng)
line.config.XTRACK_FRINGE_FROM_PTC = True
R_ptc = line.compute_one_turn_matrix_finite_differences(particle_on_co=p0.copy())
line.track(p_ptc)

assert np.isclose(p_ng.x, p_ptc.x, rtol=0, atol=1e-10)
assert np.isclose(p_ng.px, p_ptc.px, rtol=0, atol=1e-12)
assert np.isclose(p_ng.y, p_ptc.y, rtol=0, atol=1e-12)
assert np.isclose(p_ng.py, p_ptc.py, rtol=0, atol=1e-12)
assert np.isclose(p_ng.delta, p_ptc.delta, rtol=0, atol=1e-12)
assert np.isclose(p_ng.s, p_ptc.s, rtol=0, atol=1e-12)
assert np.isclose(p_ng.zeta, p_ptc.zeta, rtol=0, atol=1e-10)


import xtrack as xt
import numpy as np
import time

line = xt.Line.from_json('../../test_data/fcc_ee/fccee_h_thin.json')

tt = line.get_table()

tt_entry_quads = tt.rows['q.*\.\.0']

insertions = []
line.env.new('corrector', xt.Multipole, knl=[0])
for ii, name_quad in enumerate(tt_entry_quads.name):
    if ii%20 == 0:
        print(f'Processing {ii}/{len(tt_entry_quads)}')
    insertions.append(line.env.new(f'bpm_{ii}', xt.Marker, at=name_quad+'@start'))
    insertions.append(line.env.new(f'corrector_{ii}', 'corrector')) # after bpm
t1 = time.time()
line.insert(insertions)
t2 = time.time()
print(f't_insert = {t2-t1:.2f} s')

tt = line.get_table()
tt_bpm = tt.rows['bpm.*']

tt_h_correctors = tt.rows['corrector.*']
line.steering_correctors_x = tt_h_correctors.name
tt_v_correctors = tt.rows['corrector.*']
line.steering_correctors_y = tt_v_correctors.name

tt_monitors = tt.rows['bpm.*']
line.steering_monitors_x = tt_monitors.name
line.steering_monitors_y = tt_monitors.name

tw_ref = line.twiss4d()

orbit_correction = line.correct_trajectory(twiss_table=tw_ref, run=False)

# Inspect singular values of the response matrices
x_sv = orbit_correction.x_correction.singular_values
y_sv = orbit_correction.y_correction.singular_values

line['qc1r1.1'].knl[0] = 1e-6
line['qc1r1.1'].num_multipole_kicks = 1

t0 = time.time()
orbit_correction.correct()
t1 = time.time()
print(f'Correction time: {t1-t0} s')

# import matplotlib.pyplot as plt
# plt.close('all')

# fig2 = plt.figure(1)
# plt.semilogy(np.abs(x_sv), '.-', label='x')
# plt.semilogy(np.abs(y_sv), '.-', label='y')
# plt.legend()
# plt.xlabel('mode')
# plt.ylabel('singular value (modulus)')

# plt.show()



from cpymad.madx import Madx
import numpy as np

import xtrack as xt

class ActionSeparatrix(xt.Action):

    def __init__(self, line, n_test=20, range_test=(0, 1e-2),
                             range_fit=(0.015, 0.025),
                             i_part_fit=None,
                             num_turns=2000,
                             x_spiral_meas=None):
        self.line = line
        self.n_test = n_test
        self.range_test = range_test
        self.range_fit = range_fit
        self.i_part_fit = i_part_fit
        self.num_turns = num_turns
        self.x_spiral_meas = x_spiral_meas

    def run(self):
        line = self.line
        tw = line.twiss(method='4d')

        p_test = line.build_particles(x=np.linspace(self.range_test[0],
                                                    self.range_test[1],
                                                    self.n_test), px=0)
        line.track(p_test, num_turns=self.num_turns, turn_by_turn_monitor=True)
        mon_test = line.record_last_track
        norm_coord_test = tw.get_normalized_coordinates(mon_test)

        message = 'all ok'
        try:
            i_first_unstable = np.where(mon_test.x.max(axis=1) > 0.02)[0][0] # first unstable particle
            if self.i_part_fit is None:
                i_part = i_first_unstable # first unstable particle
            else:
                i_part = self.i_part_fit
            # i_part = len(p_test.x) - 1

            x_t = mon_test.x[i_part, :]
            px_t = mon_test.px[i_part, :]
            x_norm_t = norm_coord_test.x_norm[i_part, :]
            px_norm_t = norm_coord_test.px_norm[i_part, :]

            # Select branch closer to the x-axis
            # mask_branch = (x_norm_t > 0) & (px_norm_t > -2 * x_norm_t) & (px_norm_t < 2 * x_norm_t)
            mask_branch = (x_norm_t > 0) & (px_norm_t > 0)
            if not mask_branch.any():
                message = 'no branch'
            x_branch = x_t[mask_branch]
            px_branch = px_t[mask_branch]
            x_norm_branch = x_norm_t[mask_branch]
            px_norm_branch = px_norm_t[mask_branch]

            # for when separatrix makes a closed loop
            i_first_loop = np.where(x_branch>self.range_fit[1])[0][0]

            mask_fit = (x_branch > self.range_fit[0]) & (x_branch < self.range_fit[1])
            mask_fit[i_first_loop:] = False
            if not mask_fit.any():
                message = 'no fit'
            poly_geom = np.polyfit(x_branch[mask_fit], px_branch[mask_fit], 1)
            poly_norm = np.polyfit(x_norm_branch[mask_fit], px_norm_branch[mask_fit], 1)

            r_sep_norm = np.abs(poly_norm[1]) / np.sqrt(poly_norm[0]**2 + 1)

            if self.x_spiral_meas is not None:
                x_first_unstable = mon_test.x[i_first_unstable, :]
                i_closest = np.argmin(np.abs(x_first_unstable - self.x_spiral_meas))
                dx_spiral = (mon_test.x[i_first_unstable, i_closest + 3]
                             - mon_test.x[i_first_unstable, i_closest])
                dpx_spiral = (mon_test.px[i_first_unstable, i_closest + 3]
                             - mon_test.px[i_first_unstable, i_closest])
                djx_spiral = (
                    norm_coord_test.x_norm[i_first_unstable, i_closest + 3]**2
                  + norm_coord_test.px_norm[i_first_unstable, i_closest + 3]**2
                  - norm_coord_test.x_norm[i_first_unstable, i_closest]**2
                  - norm_coord_test.px_norm[i_first_unstable, i_closest]**2)
                slope_norm_spiral = (norm_coord_test.px_norm[i_first_unstable, i_closest]
                                     / norm_coord_test.x_norm[i_first_unstable, i_closest])
            else:
                dx_spiral = 999.
                dpx_spiral = 999.
                djx_spiral = 999.
                slope_norm_spiral = 999.

            out = {
                'poly_geom': poly_geom,
                'poly_norm': poly_norm,
                'r_sep_norm': r_sep_norm,
                'slope_geom': poly_geom[0],
                'slope_norm': poly_norm[0],
                'n_fit': np.sum(mask_fit),
                'mon' : mon_test,
                'norm_coord': norm_coord_test,
                'i_part': i_part,
                'message': message,
                'x_fit': x_branch[mask_fit],
                'px_fit': px_branch[mask_fit],
                'dx_spiral': dx_spiral,
                'dpx_spiral': dpx_spiral,
                'djx_spiral': djx_spiral,
                'slope_norm_spiral': slope_norm_spiral,
            }
        except Exception as err:
            print(err)
            out = {
                'poly_geom': [0, 0],
                'poly_norm': [0, 0],
                'r_sep_norm': 999.,
                'slope_geom': 999.,
                'slope_norm': 999.,
                'n_fit': 0,
                'mon' : mon_test,
                'norm_coord': norm_coord_test,
                'i_part': 0,
                'message': message,
            }

        return out

def plot_res(res, title=None):
    mon = res['mon']
    norm_coord = res['norm_coord']

    plt.figure(figsize=(10, 5))
    ax_geom = plt.subplot(1, 2, 1)
    plt.plot(mon.x.T, mon.px.T, '.', markersize=1)
    # plt.plot(x_branch, px_branch, '.k', markersize=3)

    plt.ylabel(r'$p_x$')
    plt.xlabel(r'$x$ [m]')

    ax_norm = plt.subplot(1, 2, 2)
    plt.plot(norm_coord.x_norm.T, norm_coord.px_norm.T, '.', markersize=1)
    # plt.plot(x_norm_branch, px_norm_branch, '.k', markersize=3)
    plt.axis('equal')

    poly_geom = res['poly_geom']
    poly_norm = res['poly_norm']
    x_fit_geom = np.linspace(-0.2, 0.2, 10)
    px_fit_geom = poly_geom[0] * x_fit_geom + poly_geom[1]
    x_fit_norm = np.linspace(-0.2, 0.2, 10)
    px_fit_norm = poly_norm[0] * x_fit_norm + poly_norm[1]

    ax_geom.plot(x_fit_geom, px_fit_geom, 'grey')
    ax_norm.plot(x_fit_norm, px_fit_norm, 'grey')

    if title is not None:
        plt.suptitle(title)

# -----------------------------------------------------------------------------

test_data_folder = '../../test_data/'
mad = Madx()

mad.call(test_data_folder + 'pimms/PIMM.seq')
mad.call(test_data_folder + 'pimms/betatron.str')
mad.beam(particle='proton', gamma=1.21315778) # 200 MeV
mad.use('pimms')
seq = mad.sequence.pimms
def_expr = True

line = xt.Line.from_madx_sequence(seq, deferred_expressions=def_expr)
line.particle_ref = xt.Particles(gamma0=seq.beam.gamma,
                                 mass0=seq.beam.mass * 1e9,
                                 q0=seq.beam.charge)
line.configure_bend_model(core='full', edge='full')


line.insert_element(
            'septum',
            xt.LimitRect(min_x=-0.1, max_x=0.1, min_y=-0.1, max_y=0.1),
            index='pimms_start')


tw0 = line.twiss(method='4d')

line.vars['k2xrr_a_extr'] = 0
line.vars['k2xrr_b_extr'] = 0
line.vars['k2xrr_a'] = line.vars['k2xrr_a_extr']
line.vars['k2xrr_b'] = line.vars['k2xrr_b_extr']

optq = line.match(
    solve=False,
    method='4d',
    vary=[
        xt.VaryList(['qf1k1', 'qd1k1', 'qf2k1'], step=1e-3, tag='quad'),
        xt.VaryList(['k2xcf', 'k2xcd'], step=1e-3, tag='sext'),
    ],
    targets=[
        xt.TargetSet(qx=1.661, qy=1.72, tol=1e-6, tag='tunes'),
        xt.TargetSet(dqx=-0.1, dqy=-0.1, tol=1e-3, tag="chrom"),
        xt.Target(dx=0, tol=1e-3, at='es', tag='disp'),
    ]
)
optq.disable_targets(tag='chrom')
optq.disable_vary(tag='sext')
optq.solve()
optq.enable_all_targets()
optq.enable_all_vary()
optq.solve()

tw1 = line.twiss(method='4d')

line.vars['k2xrr_a_extr'] = 0
line.vars['k2xrr_b_extr'] = 7.5
tw2 = line.twiss(method='4d')

import matplotlib.pyplot as plt
plt.close('all')
plt.figure(33)
ax1 = plt.subplot(2, 1, 1)
plt.plot(tw0.s, tw0.betx, '.-')
plt.plot(tw1.s, tw1.betx, '.-')
plt.plot(tw0.s, tw0.bety, '.-')
plt.plot(tw1.s, tw1.bety, '.-')
plt.ylabel(r'$\beta$ [m]')

ax2=plt.subplot(2, 1, 2, sharex=ax1)
plt.plot(tw0.s, tw0.dx, '.-')
plt.plot(tw1.s, tw1.dx, '.-')

act_show= ActionSeparatrix(line, range_test=(0e-3, 2e-2), range_fit=(2e-2, 3e-2),
                                n_test=30, x_spiral_meas=0.02)
res0 = act_show.run()



# Tune closer to resonance for separatrix matching
optq.targets[0].value = 1.667
optq.solve()

# plt.axvline(x=tw2['s', 'xrr'], color='green', linestyle='--')

act_match = ActionSeparatrix(line, range_test=(0e-3, 2e-3), range_fit=(2.5e-3, 3.e-3),
                                n_test=5, i_part_fit=4)
res_m0 = act_match.run()

opt = line.match(
    solve=False,
    method='4d',
    vary=xt.VaryList(['k2xrr_a_extr', 'k2xrr_b_extr'], step=1e-3, tag='resonance'),
    targets=[
        act_match.target('r_sep_norm', res_m0['r_sep_norm'], tol=2e-5, tag='resonance', weight=1e4),
        act_match.target('slope_norm', res_m0['slope_norm'], tol=0.05, tag='resonance'),
    ]
)

while opt.targets[1].value > 0.3:
    opt.targets[1].value -= 0.02
    opt.step(40)
    opt.target_status()
    opt.vary_status()

res_m1 = act_match.run()


# Tune closer to resonance for separatrix matching
optq.targets[0].value = 1.661
optq.solve()

res1 = act_show.run()

plot_res(res0, 'before')
plot_res(res1, 'after')
plot_res(res_m0, 'match - first point')
plot_res(res_m1, 'match - last point')

tw = line.twiss(method='4d')

num_particles = 5000
x_norm = np.random.normal(size=num_particles)
px_norm = np.random.normal(size=num_particles)
y_norm = np.random.normal(size=num_particles)
py_norm = np.random.normal(size=num_particles)
delta = 5e-4 * np.random.normal(size=num_particles)
particles = line.build_particles(
    weight=1e10/num_particles,
    method='4d',
    nemitt_x=1e-6, nemitt_y=1e-6,
    x_norm=x_norm, px_norm=px_norm, y_norm=y_norm, py_norm=py_norm,
    delta=delta)
tab = tw.get_normalized_coordinates(particles)




# plt.figure(100)
# plt.plot(x_fit_geom, px_fit_geom, 'grey')
# plt.plot(particles.x, particles.px, '.', markersize=1)
# plt.ylabel(r'$p_x$')
# plt.xlabel(r'$x$ [m]')

# plt.figure(101)
# plt.plot(x_fit_norm, px_fit_norm, 'grey')
# plt.plot(tab.x_norm, tab.px_norm, '.', markersize=1)
# plt.axis('equal')

line.discard_tracker()

class SpillExcitation:
    def __init__(self):
        self.intensity = []
        self.amplitude = 1e-6
        self.gain = 0.
        self.amplitude_max = 100e-6
        self.target_rate = 0.98e10/ 15000
        self.n_ave = 100
        self._i_turn = 0

        self._amplitude_log = []
        self._rate_log = []
        self._gain_log = []

    def track(self, p):
        self.intensity.append(np.sum(p.weight[p.state > 0]))
        i_turn_0 = self._i_turn - self.n_ave
        if i_turn_0 < 0:
            i_turn_0 = 0
        rate = -(self.intensity[self._i_turn] - self.intensity[i_turn_0]) / (self._i_turn - i_turn_0)
        if self._i_turn > 10:
            self.amplitude -= self.amplitude * self.gain * (rate - self.target_rate)/self.target_rate
        if self.amplitude > self.amplitude_max:
            self.amplitude = self.amplitude_max
        if self.amplitude < 0:
            self.amplitude = 1e-7
        self._amplitude_log.append(self.amplitude)
        self._rate_log.append(rate)
        self._gain_log.append(self.gain)
        p.px[p.state > 0] += self.amplitude * np.random.normal(size=np.sum(p.state > 0))
        self._i_turn += 1

line.insert_element('spill_exc', SpillExcitation(), at_s=0)
import xobjects as xo
line.build_tracker(_context=xo.ContextCpu('auto'))

line.functions['fun_xsext'] = xt.FunctionPieceWiseLinear(x=[0, 0.5e-3], y=[0, 1.])
line.vars['k2xrr_a'] = line.vars['k2xrr_a_extr'] * line.functions['fun_xsext'](line.vars['t_turn_s'])
line.vars['k2xrr_b'] = line.vars['k2xrr_b_extr'] * line.functions['fun_xsext'](line.vars['t_turn_s'])

line.functions['fun_gain'] = xt.FunctionPieceWiseLinear(x=[0, 0.25e-3, 0.5e-3], y=[0, 0, .001])
line.vars['gain'] = line.functions['fun_gain'](line.vars['t_turn_s'])
line.element_refs['spill_exc'].gain = line.vars['gain']

line['septum'].max_x = 0.02

line.enable_time_dependent_vars = True
line.track(particles, num_turns=15000, with_progress=True)

plt.figure(1000)
ax1 = plt.subplot(4,1,1)
plt.plot(line['spill_exc']._amplitude_log)

ax2 = plt.subplot(4,1,2, sharex=ax1)
plt.plot(line['spill_exc'].intensity)
plt.ylim(bottom=0)

ax3 = plt.subplot(4,1,3, sharex=ax1)
plt.plot(line['spill_exc']._rate_log)
plt.axhline(line['spill_exc'].target_rate, color='grey')

ax4 = plt.subplot(4,1,4, sharex=ax1)
ax4b = ax4.twinx()
plt.plot(line['spill_exc']._gain_log)

plt.figure(1001)
plt.plot(particles.x, particles.px, '.', markersize=2)
# plt.plot(x_fit_geom, px_fit_geom, 'grey')
plt.show()


import json

import xtrack as xt
import xobjects as xo

with open('../../test_data/hllhc14_no_errors_with_coupling_knobs/line_b1.json',
            'r') as fid:
    dct_b1 = json.load(fid)
line = xt.Line.from_dict(dct_b1)

context = xo.ContextCpu()
# context = xo.ContextCupy()
# context = xo.ContextPyopencl()

line.build_tracker(_context=context)
target_qx = 62.315
target_qy = 60.325

tw = line.twiss()

# Try to measure and match coupling
line.vars['cmrskew'] = 1e-3
line.vars['cmiskew'] = 1e-3

# Match coupling
line.match(
    vary=[
        xt.Vary(name='cmrskew', limits=[-0.5e-2, 0.5e-2], step=1e-5),
        xt.Vary(name='cmiskew', limits=[-0.5e-2, 0.5e-2], step=1e-5)],
    targets=[
        # xt.Target('c_r1_avg', 0, tol=2e-2),
        # xt.Target('c_r2_avg', 0, tol=2e-2),
        xt.Target('c_minus', 0, tol=1e-4),
        ])



import json
import pathlib
import time
from scipy.optimize import minimize

import xobjects as xo
import xpart as xp
import xtrack as xt
from xobjects.test_helpers import for_all_test_contexts

test_data_folder = pathlib.Path(
            __file__).parent.joinpath('../test_data').absolute()
path_line = test_data_folder.joinpath(
                'hllhc15_noerrors_nobb/line_w_knobs_and_particle.json')



@for_all_test_contexts
def test_match_tune_chromaticity(test_context):

    with open(path_line) as f:
        dct = json.load(f)

    line = xt.Line.from_dict(dct['line'])
    line.particle_ref = xp.Particles.from_dict(dct['particle'])

    line.build_tracker(_context=test_context)

    print('\nInitial twiss parameters')
    tw_before = line.twiss()
    print(f"Qx = {tw_before['qx']:.5f} Qy = {tw_before['qy']:.5f} "
        f"Q'x = {tw_before['dqx']:.5f} Q'y = {tw_before['dqy']:.5f}")

    print(f"kqtf.b1 = {line.vars['kqtf.b1']._value}")
    print(f"kqtd.b1 = {line.vars['kqtd.b1']._value}")
    print(f"ksf.b1 = {line.vars['ksf.b1']._value}")
    print(f"ksd.b1 = {line.vars['ksd.b1']._value}")

    t1 = time.time()
    line.match(
        vary=[
            xt.Vary('kqtf.b1', step=1e-8),
            xt.Vary('kqtd.b1', step=1e-8),
            xt.Vary('ksf.b1', step=1e-8),
            xt.Vary('ksd.b1', step=1e-8),
        ],
        targets = [
            xt.Target('qx', 62.315, tol=1e-5),
            xt.Target(lambda tw: tw['qx'] - tw['qy'], 1.99, tol=1e-5),
            xt.Target('dqx', 10.0, tol=0.05),
            xt.Target('dqy', 12.0, tol=0.05)])
    t2 = time.time()
    print('\nTime fsolve: ', t2-t1)

    tw_final = line.twiss()
    print('\nFinal twiss parameters')
    print(f"Qx = {tw_final['qx']:.5f} Qy = {tw_final['qy']:.5f} "
        f"Q'x = {tw_final['dqx']:.5f} Q'y = {tw_final['dqy']:.5f}")
    print(f"kqtf.b1 = {line.vars['kqtf.b1']._value}")
    print(f"kqtd.b1 = {line.vars['kqtd.b1']._value}")
    print(f"ksf.b1 = {line.vars['ksf.b1']._value}")
    print(f"ksd.b1 = {line.vars['ksd.b1']._value}")

    xo.assert_allclose(tw_final['qx'], 62.315, atol=1e-5)
    xo.assert_allclose(tw_final['qy'], 60.325, atol=1e-5)
    xo.assert_allclose(tw_final['dqx'], 10.0, atol=0.05)
    xo.assert_allclose(tw_final['dqy'], 12.0, atol=0.05)


    t1 = time.time()
    line.match(
        vary=[
            xt.Vary('kqtf.b1', step=1e-8),
            xt.Vary('kqtd.b1', step=1e-8),
            xt.Vary('ksf.b1', step=1e-8),
            xt.Vary('ksd.b1', step=1e-8),
        ],
        targets = [
            xt.Target('qx', 62.27, tol=1e-5),
            xt.Target('qy', 60.28, tol=1e-5),
            xt.Target('dqx', -5.0, tol=0.05),
            xt.Target('dqy', -7.0, tol=0.05),])
    t2 = time.time()
    print('\nTime fsolve: ', t2-t1)

    tw_final = line.twiss()
    print('\nFinal twiss parameters')
    print(f"Qx = {tw_final['qx']:.5f} Qy = {tw_final['qy']:.5f} "
        f"Q'x = {tw_final['dqx']:.5f} Q'y = {tw_final['dqy']:.5f}")
    print(f"kqtf.b1 = {line.vars['kqtf.b1']._value}")
    print(f"kqtd.b1 = {line.vars['kqtd.b1']._value}")
    print(f"ksf.b1 = {line.vars['ksf.b1']._value}")
    print(f"ksd.b1 = {line.vars['ksd.b1']._value}")

    xo.assert_allclose(tw_final['qx'], 62.27, atol=1e-5)
    xo.assert_allclose(tw_final['qy'], 60.28, atol=1e-5)
    xo.assert_allclose(tw_final['dqx'], -5.0, atol=0.05)
    xo.assert_allclose(tw_final['dqy'], -7.0, atol=0.05)

    # Trying 4d matching
    for ee in line.elements:
        if isinstance(ee, xt.Cavity):
            ee.voltage = 0.0
    line.match(method='4d', # <-- 4d matchin
        freeze_longitudinal=True,
        vary=[
            xt.Vary('kqtf.b1', step=1e-8),
            xt.Vary('kqtd.b1', step=1e-8),
            xt.Vary('ksf.b1', step=1e-8),
            xt.Vary('ksd.b1', step=1e-8),
        ],
        targets = [
            xt.Target('qx', 62.29, tol=1e-5),
            xt.Target('qy', 60.31, tol=1e-5),
            xt.Target('dqx', 6.0, tol=0.05),
            xt.Target('dqy', 4.0, tol=0.05),])
    t2 = time.time()
    print('\nTime fsolve: ', t2-t1)

    tw_final = line.twiss(method='4d')
    print('\nFinal twiss parameters')
    print(f"Qx = {tw_final['qx']:.5f} Qy = {tw_final['qy']:.5f} "
        f"Q'x = {tw_final['dqx']:.5f} Q'y = {tw_final['dqy']:.5f}")
    print(f"kqtf.b1 = {line.vars['kqtf.b1']._value}")
    print(f"kqtd.b1 = {line.vars['kqtd.b1']._value}")
    print(f"ksf.b1 = {line.vars['ksf.b1']._value}")
    print(f"ksd.b1 = {line.vars['ksd.b1']._value}")

    xo.assert_allclose(tw_final['qx'], 62.29, atol=1e-5)
    xo.assert_allclose(tw_final['qy'], 60.31, atol=1e-5)
    xo.assert_allclose(tw_final['dqx'],  6.0, atol=0.05)
    xo.assert_allclose(tw_final['dqy'],  4.0, atol=0.05)

@for_all_test_contexts
def test_match_tune_chromaticity_scalar(test_context):

    with open(path_line) as f:
        dct = json.load(f)

    line = xt.Line.from_dict(dct['line'])
    line.particle_ref = xp.Particles.from_dict(dct['particle'])

    line.build_tracker(_context=test_context)

    print('\nInitial twiss parameters')
    tw_before = line.twiss()
    print(f"Qx = {tw_before['qx']:.5f} Qy = {tw_before['qy']:.5f} "
        f"Q'x = {tw_before['dqx']:.5f} Q'y = {tw_before['dqy']:.5f}")

    print(f"kqtf.b1 = {line.vars['kqtf.b1']._value}")
    print(f"kqtd.b1 = {line.vars['kqtd.b1']._value}")
    print(f"ksf.b1 = {line.vars['ksf.b1']._value}")
    print(f"ksd.b1 = {line.vars['ksd.b1']._value}")

    t1 = time.time()
    opt = line.match(
        solve=False,
        vary=[
            xt.Vary('kqtf.b1', step=1e-8),
            xt.Vary('kqtd.b1', step=1e-8),
            xt.Vary('ksf.b1', step=1e-8),
            xt.Vary('ksd.b1', step=1e-8),
        ],
        targets = [
            xt.Target('qx', 62.315, tol=1e-6),
            xt.Target('qy', 60.325, tol=1e-6),
            xt.Target('dqx', 10.0, tol=0.05),
            xt.Target('dqy', 12.0, tol=0.05)])

    merit_function = opt.get_merit_function(return_scalar=True, check_limits=False)
    opt.check_limits = False
    bounds = [(-1e-4, 1e-4), (-1e-4, 1e-4), (-10, 10), (-10, 10)]
    x0 = merit_function.get_x()

    result = minimize(merit_function, x0=x0, bounds=bounds, jac=merit_function.get_jacobian, method='L-BFGS-B')

    t2 = time.time()
    print('\nTime fsolve: ', t2-t1)

    merit_function.set_x(result.x)

    tw_final = line.twiss()
    print('\nFinal twiss parameters')
    print(f"Qx = {tw_final['qx']:.5f} Qy = {tw_final['qy']:.5f} "
        f"Q'x = {tw_final['dqx']:.5f} Q'y = {tw_final['dqy']:.5f}")
    print(f"kqtf.b1 = {line.vars['kqtf.b1']._value}")
    print(f"kqtd.b1 = {line.vars['kqtd.b1']._value}")
    print(f"ksf.b1 = {line.vars['ksf.b1']._value}")
    print(f"ksd.b1 = {line.vars['ksd.b1']._value}")

    xo.assert_allclose(tw_final['qx'], 62.315, atol=1e-6)
    xo.assert_allclose(tw_final['qy'], 60.325, atol=1e-6)
    xo.assert_allclose(tw_final['dqx'], 10.0, atol=0.05)
    xo.assert_allclose(tw_final['dqy'], 12.0, atol=0.05)

@for_all_test_contexts
def test_match_coupling(test_context):
    with open(test_data_folder /
        'hllhc14_no_errors_with_coupling_knobs/line_b1.json', 'r') as fid:
        dct_b1 = json.load(fid)
    line = xt.Line.from_dict(dct_b1)

    line.build_tracker(_context=test_context)

    tw = line.twiss()

    assert tw.c_minus < 1e-4

    # Try to measure and match coupling
    line.vars['cmrskew'] = 1e-3
    line.vars['cmiskew'] = 1e-3

    tw = line.twiss()
    assert tw.c_minus > 2e-4

    # Match coupling
    line.match(verbose=True,
        vary=[
            xt.Vary(name='cmrskew', limits=[-1e-2, 1e-2], step=1e-5),
            xt.Vary(name='cmiskew', limits=[-1e-2, 1e-2], step=1e-5),
        ],
        targets=[
            xt.Target('c_minus', 0, tol=1e-4)])

    tw = line.twiss()
    assert tw.c_minus < 2e-4

@for_all_test_contexts
def test_match_chroma_knob(test_context):

    with open(test_data_folder /
        'hllhc15_noerrors_nobb/line_w_knobs_and_particle.json') as f:
        dct = json.load(f)

    line = xt.Line.from_dict(dct['line'])
    line.particle_ref = xp.Particles.from_dict(dct['particle'])
    line.twiss_default['method'] = '4d'
    line.twiss_default['freeze_longitudinal'] = True
    line.build_tracker(_context=test_context)

    vary=[ xt.Vary('ksf.b1', step=1e-8),  xt.Vary('ksd.b1', step=1e-8)]
    line.match(
        vary=vary,
        targets = [xt.Target('dqx', 2.0, tol=1e-6),
                xt.Target('dqy', 2.0, tol=1e-6)])

    tw0 = line.twiss()
    line.match_knob('dqx.b1',
                knob_value_start=2.0,
                knob_value_end=3.0,
                vary=[ xt.Vary('ksf.b1', step=1e-8), xt.Vary('ksd.b1', step=1e-8)],
                targets=[
                    xt.Target('dqx', 3.0, tol=1e-6),
                    xt.Target('dqy', tw0, tol=1e-6)])

    line.match_knob('dqy.b1',
                knob_value_start=2.0,
                knob_value_end=3.0,
                vary=[ xt.Vary('ksf.b1', step=1e-8), xt.Vary('ksd.b1', step=1e-8)],
                targets=[
                    xt.Target('dqx', tw0, tol=1e-6),
                    xt.Target('dqy', 3.0, tol=1e-6)])

    tw = line.twiss()
    xo.assert_allclose(tw.dqx, 2.0, atol=1e-6)
    xo.assert_allclose(tw.dqy, 2.0, atol=1e-6)

    line.vars['dqx.b1'] = 6.0
    line.vars['dqy.b1'] = 7.0

    tw = line.twiss()
    xo.assert_allclose(tw.dqx, 6.0, atol=1e-4)
    xo.assert_allclose(tw.dqy, 7.0, atol=1e-4)











import numpy as np
import xtrack as xt

line = xt.Line.from_json(
    '../../test_data/hllhc15_thick/lhc_thick_with_knobs.json')

tw = line.twiss(method='4d')

mng = line.to_madng(sequence_name='lhcb1')

mng["mytwtable", 'mytwflow'] = mng.twiss(
    sequence=mng.lhcb1, method=4, mapdef=2, implicit=True, nslice=3, save="'atbody'")


mng.send(r'''
local damap, vector in MAD
local lhc = MADX.lhcb1

-- list of octupolar RDTs
local rdts = {"f4000", "f3100", "f2020", "f1120"}

-- create phase-space damap at 4th order
local X0 = damap {nv=6, mo=4}

-- twiss with RDTs
local mtbl = track {sequence=lhc, X0=X0, info=2, observe=0, savemap=true}

mtbl:addcol("symperror", \ri -> mtbl.__map[ri]:get1():symperr())

local symp_err_vect = vector(#mtbl):fill(mtbl.symperror)

symp_err_vect:print("symperror")

-- send columns to Python
py:send({mtbl.s, symp_err_vect})

''')

s, symperror = mng.recv()

import matplotlib.pyplot as plt
plt.close('all')
plt.figure(1)
ax1 = plt.subplot(2, 1, 1)
plt.plot(tw.s, tw.betx)

ax2 = plt.subplot(2, 1, 2, sharex=ax1)
plt.plot(s, symperror)

plt.show()



import xtrack as xt
from numpy.testing import assert_allclose

def test_sign_rotx():
    pt=xt.Particles(y=1)
    xt.XRotation(30).track(pt)
    assert_allclose(pt.py[0],1/2) #anti-clockwise rotation
    xt.XRotation(-30).track(pt)
    assert_allclose(pt.py[0],0,rtol=0,atol=1e-16)
    assert_allclose(pt.y[0],1)
    assert_allclose(pt.zeta[0],0,rtol=0,atol=1e-16)

def test_sign_roty():
    pt=xt.Particles(x=1,p0c=1e20)
    xt.YRotation(30).track(pt)
    assert_allclose(pt.px[0],-1/2) #anti-clockwise rotation

def test_sign_rots():
    pt=xt.Particles(y=1)
    xt.SRotation(30).track(pt)
    assert_allclose(pt.x[0],1/2) #anti-clockwise rotation
    pt=xt.Particles(x=1)
    xt.SRotation(30).track(pt)
    assert_allclose(pt.y[0],-1/2) #anti-clockwise rotation



import xtrack as xt
import numpy as np

from scipy.constants import hbar
from scipy.constants import electron_volt
from scipy.constants import c as clight

env = xt.load_madx_lattice('b075_2024.09.25.madx')
line = env.ring
line.particle_ref = xt.Particles(energy0=2.7e9, mass0=xt.ELECTRON_MASS_EV)
line.configure_bend_model(num_multipole_kicks=20)

line['vrf'] = 1.8e6
line['frf'] = 499.6e6
line['lagrf'] = 180.

line.insert(
    env.new('cav', 'Cavity', voltage='vrf', frequency='frf', lag='lagrf', at=0))

tt = line.get_table()
tw4d_thick = line.twiss4d()
tw6d_thick = line.twiss()

env['ring_thick'] = env.ring.copy(shallow=True)

line.discard_tracker()
slicing_strategies = [
    xt.Strategy(slicing=None),  # Default
    xt.Strategy(slicing=xt.Teapot(20), element_type=xt.Bend),
    xt.Strategy(slicing=xt.Teapot(8), element_type=xt.Quadrupole),
]
line.slice_thick_elements(slicing_strategies)

tw4d = line.twiss4d()
tw6d = line.twiss()

line.configure_radiation(model='mean')

tw_rad = line.twiss(eneloss_and_damping=True, strengths=True)

# import matplotlib.pyplot as plt
# plt.close('all')
# pl = tw_rad.plot(yl='y', yr='dy')
# pl.xlim(tw_rad['s', 's.wig'] - 10, tw_rad['s', 'e.wig'] + 10)
# plt.show()

# from synchrotron_integrals import SynchrotronIntegral as synint
# integrals = synint(line)

tw = tw_rad

angle_rad = tw['angle_rad']
rot_s_rad = tw['rot_s_rad']
x = tw['x']
y = tw['y']
kin_px = tw['kin_px']
kin_py = tw['kin_py']
delta = tw['delta']
length = tw['length']

betx = tw['betx']             # Twiss beta function x
alfx = tw['alfx']             # Twiss alpha x
gamx = tw['gamx']             # Twiss gamma x
bety = tw['bety']             # Twiss beta function y
alfy = tw['alfy']             # Twiss alpha y
gamy = tw['gamy']             # Twiss gamma y
dx = tw['dx']                 # Dispersion x
dy = tw['dy']                 # Dispersion y
dpx = tw['dpx']               # Dispersion px
dpy = tw['dpy']               # Dispersion py

mass0 = line.particle_ref.mass0
r0 = line.particle_ref.get_classical_particle_radius0()
gamma0 = tw.gamma0

dxprime = dpx * (1 - delta) - kin_px
dyprime = dpy * (1 - delta) - kin_py

# Curvature of the reference trajectory
mask = length != 0
kappa0_x = 0 * angle_rad
kappa0_y = 0 * angle_rad
kappa0_x[mask] = angle_rad[mask] * np.cos(rot_s_rad[mask]) / length[mask]
kappa0_y[mask] = angle_rad[mask] * np.sin(rot_s_rad[mask]) / length[mask]
kappa0 = np.sqrt(kappa0_x**2 + kappa0_y**2)

# Field index
fieldindex = 0 * angle_rad
k1 = 0 * angle_rad
k1[mask] = tw.k1l[mask] / length[mask]
mask_k0 = kappa0 > 0
fieldindex[mask_k0] = -1. / kappa0[mask_k0]**2 * k1[mask_k0]

# Compute x', y', x'', y''
ps = np.sqrt((1 + delta)**2 - kin_px**2 - kin_py**2)
xp = kin_px / ps
yp = kin_py / ps
xp_ele = xp * 0
yp_ele = yp * 0
xp_ele[:-1] = (xp[:-1] + xp[1:]) / 2
yp_ele[:-1] = (yp[:-1] + yp[1:]) / 2

mask_length = length != 0
xpp_ele = xp_ele * 0
ypp_ele = yp_ele * 0
xpp_ele[mask_length] = np.diff(xp, append=0)[mask_length] / length[mask_length]
ypp_ele[mask_length] = np.diff(yp, append=0)[mask_length] / length[mask_length]

# Curvature of the particle trajectory
hhh = 1 + kappa0_x * x + kappa0_y * y
hprime = kappa0_x * xp_ele + kappa0_y * yp_ele
mask1 = xpp_ele**2 + hhh**2 != 0
mask2 = xpp_ele**2 + hhh**2 != 0
kappa_x = (-(hhh * (xpp_ele - hhh * kappa0_x) - 2 * hprime * xp_ele)[mask1]
           / (xp_ele**2 + hhh**2)[mask1]**(3/2))
kappa_y = (-(hhh * (ypp_ele - hhh * kappa0_y) - 2 * hprime * yp_ele)[mask2]
           / (yp_ele**2 + hhh**2)[mask2]**(3/2))

# Curly H
Hx_rad = gamx * dx**2 + 2*alfx * dx * dxprime + betx * dxprime**2
Hy_rad = gamy * dy**2 + 2*alfy * dy * dyprime + bety * dyprime**2

# Integrands
i1x_integrand = kappa_x * dx
i1y_integrand = kappa_y * dy

i2x_integrand = kappa_x * kappa_x
i2y_integrand = kappa_y * kappa_y

i3x_integrand = kappa_x * kappa_x * kappa_x
i3y_integrand = kappa_y * kappa_y * kappa_y

i4x_integrand = kappa_x * kappa_x * kappa_x * dx * (1 - 2 * fieldindex)
i4y_integrand = kappa_y * kappa_y * kappa_y * dy * (1 - 2 * fieldindex)

i5x_integrand = np.abs(kappa_x*kappa_x*kappa_x) * Hx_rad
i5y_integrand = np.abs(kappa_y*kappa_y*kappa_y) * Hy_rad

# Integrate
i1x = np.sum(i1x_integrand * length)
i1y = np.sum(i1y_integrand * length)
i2x = np.sum(i2x_integrand * length)
i2y = np.sum(i2y_integrand * length)
i3x = np.sum(i3x_integrand * length)
i3y = np.sum(i3y_integrand * length)
i4x = np.sum(i4x_integrand * length)
i4y = np.sum(i4y_integrand * length)
i5x = np.sum(i5x_integrand * length)
i5y = np.sum(i5y_integrand * length)

eq_gemitt_x = (55/(32 * 3**(1/2)) * hbar / electron_volt * clight
               / mass0 * gamma0**2 * i5x / (i2x + i2y - i4x))
eq_gemitt_y = (55/(32 * 3**(1/2)) * hbar / electron_volt * clight
               / mass0 * gamma0**2 * i5y / (i2x + i2y - i4y))

damping_constant_x_s = r0/3 * gamma0**3 * clight/tw.circumference * (i2x + i2y - i4x)
damping_constant_y_s = r0/3 * gamma0**3 * clight/tw.circumference * (i2x + i2y - i4y)
damping_constant_zeta_s = r0/3 * gamma0**3 * clight/tw.circumference * (2 * (i2x + i2y) + i4x + i4y)

tw_integrals = line.twiss(radiation_integrals=True)



tw_rad = line.twiss(eneloss_and_damping=True, strengths=True)

print('ex rad int:', tw_integrals.rad_int_eq_gemitt_x)
print('ex Chao:   ', tw_rad.eq_gemitt_x)
print('ey rad int:', tw_integrals.rad_int_eq_gemitt_y)
print('ey Chao:   ', tw_rad.eq_gemitt_y)

print('damping rate x [s^-1] rad int:   ', tw_integrals.rad_int_damping_constant_x_s)
print('damping rate x [s^-1] eigenval:  ', tw_rad.damping_constants_s[0])
print('damping rate y [s^-1] rad int:   ', tw_integrals.rad_int_damping_constant_y_s)
print('damping rate y [s^-1] eigenval:  ', tw_rad.damping_constants_s[1])
print('damping rate z [s^-1] rad int:   ', tw_integrals.rad_int_damping_constant_zeta_s)
print('damping rate z [s^-1] eigenval:  ', tw_rad.damping_constants_s[2])


import time

import xtrack as xt

import xtrack._temp.lhc_match as lm

def second_order_chromaticity(line, ddelta=1e-5, **kwargs):

    tw0 = line.twiss(**kwargs)

    kwargs.pop('delta0', None)

    tw_plus = line.twiss(delta0=tw0.delta[0] + ddelta, **kwargs)
    tw_minus = line.twiss(delta0=tw0.delta[0] - ddelta, **kwargs)

    ddqx = (tw_plus.dqx - tw_minus.dqx) / (2 * ddelta)
    ddqy = (tw_plus.dqy - tw_minus.dqy) / (2 * ddelta)

    return ddqx, ddqy


default_tol = {None: 1e-8, 'betx': 1e-6, 'bety': 1e-6} # to have no rematching w.r.t. madx

collider = xt.Environment.from_json(
    "../../test_data/hllhc15_thick/hllhc15_collider_thick.json")
collider.build_trackers()
collider.vars.load_madx_optics_file(
    "../../test_data/hllhc15_thick/opt_round_150_1500.madx")

c0 = collider.copy()

lm.set_var_limits_and_steps(collider)

tw0 = collider.twiss()
for ll in collider.line_names:
    line = collider[ll]
    ddqx, ddqy = second_order_chromaticity(line)
    tw0[ll]._data['ddqx'] = ddqx
    tw0[ll]._data['ddqy'] = ddqy

beta15 = {
    'betx_ip1': 0.15,
    'bety_ip1': 0.10,
    'betx_ip5': 0.10,
    'bety_ip5': 0.15,
}

staged_match = False

optimizers = {}

for vv in collider.vars.vary_default:
    lims = collider.vars.vary_default[vv]['limits']
    if lims is None: continue
    a = lims[0]
    b = lims[1]

    if a>0 and b>0:
        collider.vars.vary_default[vv]['limits'] = (0.9*a, 1.1*b)
    elif a<0 and b<0:
        collider.vars.vary_default[vv]['limits'] = (1.1*a, 0.9*b)
    else:
        collider.vars.vary_default[vv]['limits'] = (1.1*a, 1.1*b)

arc_periodic_solution = lm.get_arc_periodic_solution(collider)

optimizers.update({'b1':{}, 'b2':{}})

t_match_irs = {}
for bn in ['b1', 'b2']:

    tt_1 = time.time()
    line_name = f'lhc{bn}'

    muxip1_l = collider.varval[f'muxip1{bn}_l']
    muyip1_l = collider.varval[f'muyip1{bn}_l']
    muxip1_r = collider.varval[f'muxip1{bn}_r']
    muyip1_r = collider.varval[f'muyip1{bn}_r']

    muxip5_l = collider.varval[f'muxip5{bn}_l']
    muyip5_l = collider.varval[f'muyip5{bn}_l']
    muxip5_r = collider.varval[f'muxip5{bn}_r']
    muyip5_r = collider.varval[f'muyip5{bn}_r']

    muxip2 = collider.varval[f'muxip2{bn}']
    muyip2 = collider.varval[f'muyip2{bn}']
    muxip4 = collider.varval[f'muxip4{bn}']
    muyip4 = collider.varval[f'muyip4{bn}']
    muxip6 = collider.varval[f'muxip6{bn}']
    muyip6 = collider.varval[f'muyip6{bn}']
    muxip8 = collider.varval[f'muxip8{bn}']
    muyip8 = collider.varval[f'muyip8{bn}']

    mux12 = collider.varval[f'mux12{bn}']
    muy12 = collider.varval[f'muy12{bn}']
    mux45 = collider.varval[f'mux45{bn}']
    muy45 = collider.varval[f'muy45{bn}']
    mux56 = collider.varval[f'mux56{bn}']
    muy56 = collider.varval[f'muy56{bn}']
    mux81 = collider.varval[f'mux81{bn}']
    muy81 = collider.varval[f'muy81{bn}']

    betx_ip1_init = collider.varval[f'betxip1{bn}']
    bety_ip1_init = collider.varval[f'betyip1{bn}']
    betx_ip5_init = collider.varval[f'betxip5{bn}']
    bety_ip5_init = collider.varval[f'betyip5{bn}']

    betx_ip2 = collider.varval[f'betxip2{bn}']
    bety_ip2 = collider.varval[f'betyip2{bn}']

    alfx_ip3 = collider.varval[f'alfxip3{bn}']
    alfy_ip3 = collider.varval[f'alfyip3{bn}']
    betx_ip3 = collider.varval[f'betxip3{bn}']
    bety_ip3 = collider.varval[f'betyip3{bn}']
    dx_ip3 = collider.varval[f'dxip3{bn}']
    dpx_ip3 = collider.varval[f'dpxip3{bn}']
    mux_ir3 = collider.varval[f'muxip3{bn}']
    muy_ir3 = collider.varval[f'muyip3{bn}']

    alfx_ip4 = collider.varval[f'alfxip4{bn}']
    alfy_ip4 = collider.varval[f'alfyip4{bn}']
    betx_ip4 = collider.varval[f'betxip4{bn}']
    bety_ip4 = collider.varval[f'betyip4{bn}']
    dx_ip4 = collider.varval[f'dxip4{bn}']
    dpx_ip4 = collider.varval[f'dpxip4{bn}']

    alfx_ip6 = collider.varval[f'alfxip6{bn}']
    alfy_ip6 = collider.varval[f'alfyip6{bn}']
    betx_ip6 = collider.varval[f'betxip6{bn}']
    bety_ip6 = collider.varval[f'betyip6{bn}']
    dx_ip6 = collider.varval[f'dxip6{bn}']
    dpx_ip6 = collider.varval[f'dpxip6{bn}']

    alfx_ip7 = collider.varval[f'alfxip7{bn}']
    alfy_ip7 = collider.varval[f'alfyip7{bn}']
    betx_ip7 = collider.varval[f'betxip7{bn}']
    bety_ip7 = collider.varval[f'betyip7{bn}']
    dx_ip7 = collider.varval[f'dxip7{bn}']
    dpx_ip7 = collider.varval[f'dpxip7{bn}']
    mux_ir7 = collider.varval[f'muxip7{bn}']
    muy_ir7 = collider.varval[f'muyip7{bn}']

    alfx_ip8 = collider.varval[f'alfxip8{bn}']
    alfy_ip8 = collider.varval[f'alfyip8{bn}']
    betx_ip8 = collider.varval[f'betxip8{bn}']
    bety_ip8 = collider.varval[f'betyip8{bn}']
    dx_ip8 = collider.varval[f'dxip8{bn}']
    dpx_ip8 = collider.varval[f'dpxip8{bn}']

    betx_ip1_target = beta15[f'betx_ip1']
    bety_ip1_target = beta15[f'bety_ip1']
    betx_ip5_target = beta15[f'betx_ip5']
    bety_ip5_target = beta15[f'bety_ip5']

    betx1_diff = betx_ip1_target - betx_ip1_init
    bety1_diff = bety_ip1_target - bety_ip1_init
    betx5_diff = betx_ip5_target - betx_ip5_init
    bety5_diff = bety_ip5_target - bety_ip5_init

    max_bet_diff = max(abs(betx1_diff), abs(bety1_diff),
                          abs(betx5_diff), abs(bety5_diff))

    n_steps = int(max_bet_diff / 0.02) + 1

    for i_step in range(n_steps):

        print(f"Step {i_step+1}/{n_steps}")
        betx_ip1 = betx_ip1_init + betx1_diff * (i_step+1) / n_steps
        bety_ip1 = bety_ip1_init + bety1_diff * (i_step+1) / n_steps
        betx_ip5 = betx_ip5_init + betx5_diff * (i_step+1) / n_steps
        bety_ip5 = bety_ip5_init + bety5_diff * (i_step+1) / n_steps

        print(f"Matching betx_ip1={betx_ip1} bety_ip1={bety_ip1} "
                f"betx_ip5={betx_ip5} bety_ip5={bety_ip5}")

        tw_sq_a81_ip1_a12 = lm.propagate_optics_from_beta_star(collider, ip_name='ip1',
                line_name=f'lhc{bn}', start=f's.ds.r8.{bn}', end=f'e.ds.l2.{bn}',
                beta_star_x=betx_ip1, beta_star_y=bety_ip1)

        tw_sq_a45_ip5_a56 = lm.propagate_optics_from_beta_star(collider, ip_name='ip5',
                line_name=f'lhc{bn}', start=f's.ds.r4.{bn}', end=f'e.ds.l6.{bn}',
                beta_star_x=betx_ip5, beta_star_y=bety_ip5)

        (mux_ir2_target, muy_ir2_target, mux_ir4_target, muy_ir4_target,
        mux_ir6_target, muy_ir6_target, mux_ir8_target, muy_ir8_target
            ) = lm.compute_ats_phase_advances_for_auxiliary_irs(line_name,
                tw_sq_a81_ip1_a12, tw_sq_a45_ip5_a56,
                muxip1_l, muyip1_l, muxip1_r, muyip1_r,
                muxip5_l, muyip5_l, muxip5_r, muyip5_r,
                muxip2, muyip2, muxip4, muyip4, muxip6, muyip6, muxip8, muyip8,
                mux12, muy12, mux45, muy45, mux56, muy56, mux81, muy81)

        print(f"Matching IR2 {bn}")
        opt = lm.rematch_ir2(collider, line_name=f'lhc{bn}',
                    boundary_conditions_left=tw_sq_a81_ip1_a12,
                    boundary_conditions_right=arc_periodic_solution[f'lhc{bn}']['23'],
                    mux_ir2=mux_ir2_target, muy_ir2=muy_ir2_target,
                    betx_ip2=betx_ip2, bety_ip2=bety_ip2,
                    solve=False, staged_match=staged_match,
                    default_tol=default_tol)
        opt.solver.max_rel_penalty_increase = 2
        opt.solve()
        optimizers[bn]['ir2'] = opt

        print(f"Matching IR3 {bn}")
        opt = lm.rematch_ir3(collider=collider, line_name=f'lhc{bn}',
                    boundary_conditions_left=arc_periodic_solution[f'lhc{bn}']['23'],
                    boundary_conditions_right=arc_periodic_solution[f'lhc{bn}']['34'],
                    mux_ir3=mux_ir3, muy_ir3=muy_ir3,
                    alfx_ip3=alfx_ip3, alfy_ip3=alfy_ip3,
                    betx_ip3=betx_ip3, bety_ip3=bety_ip3,
                    dx_ip3=dx_ip3, dpx_ip3=dpx_ip3,
                    solve=False, staged_match=staged_match, default_tol=default_tol)
        opt.solver.max_rel_penalty_increase = 2
        opt.solve()

        optimizers[bn]['ir3'] = opt

        print(f"Matching IR4 {bn}")
        opt = lm.rematch_ir4(collider=collider, line_name=f'lhc{bn}',
                    boundary_conditions_left=arc_periodic_solution[f'lhc{bn}']['34'],
                    boundary_conditions_right=tw_sq_a45_ip5_a56,
                    mux_ir4=mux_ir4_target, muy_ir4=muy_ir4_target,
                    alfx_ip4=alfx_ip4, alfy_ip4=alfy_ip4,
                    betx_ip4=betx_ip4, bety_ip4=bety_ip4,
                    dx_ip4=dx_ip4, dpx_ip4=dpx_ip4,
                    solve=False, staged_match=staged_match, default_tol=default_tol)
        opt.solver.max_rel_penalty_increase = 2
        opt.solve()
        optimizers[bn]['ir4'] = opt

        print(f"Matching IP6 {bn}")
        opt = lm.rematch_ir6(collider=collider, line_name=f'lhc{bn}',
                    boundary_conditions_left=tw_sq_a45_ip5_a56,
                    boundary_conditions_right=arc_periodic_solution[f'lhc{bn}']['67'],
                    mux_ir6=mux_ir6_target, muy_ir6=muy_ir6_target,
                    alfx_ip6=alfx_ip6, alfy_ip6=alfy_ip6,
                    betx_ip6=betx_ip6, bety_ip6=bety_ip6,
                    dx_ip6=dx_ip6, dpx_ip6=dpx_ip6,
                    solve=False, staged_match=staged_match, default_tol=default_tol)
        opt.targets[0].active = False
        opt.targets[1].active = False
        opt.targets[2].active = False
        opt.targets[3].active = False
        opt.solver.max_rel_penalty_increase = 2
        opt.solve()
        optimizers[bn]['ir6'] = opt

        print(f"Matching IP7 {bn}")
        opt = lm.rematch_ir7(collider=collider, line_name=f'lhc{bn}',
                boundary_conditions_left=arc_periodic_solution[f'lhc{bn}']['67'],
                boundary_conditions_right=arc_periodic_solution[f'lhc{bn}']['78'],
                mux_ir7=mux_ir7, muy_ir7=muy_ir7,
                alfx_ip7=alfx_ip7, alfy_ip7=alfy_ip7,
                betx_ip7=betx_ip7, bety_ip7=bety_ip7,
                dx_ip7=dx_ip7, dpx_ip7=dpx_ip7,
                solve=False, staged_match=staged_match, default_tol=default_tol)
        opt.solver.max_rel_penalty_increase = 2
        opt.solve()
        optimizers[bn]['ir7'] = opt

        print(f"Matching IP8 {bn}")
        opt = lm.rematch_ir8(collider=collider, line_name=f'lhc{bn}',
                boundary_conditions_left=arc_periodic_solution[f'lhc{bn}']['78'],
                boundary_conditions_right=tw_sq_a81_ip1_a12,
                mux_ir8=mux_ir8_target, muy_ir8=muy_ir8_target,
                alfx_ip8=alfx_ip8, alfy_ip8=alfy_ip8,
                betx_ip8=betx_ip8, bety_ip8=bety_ip8,
                dx_ip8=dx_ip8, dpx_ip8=dpx_ip8,
                solve=False, staged_match=staged_match, default_tol=default_tol)
        opt.solver.max_rel_penalty_increase = 2
        opt.solve()
        optimizers[bn]['ir8'] = opt

    tt_2 = time.time()
    t_match_irs[bn] = tt_2 - tt_1

t3 = time.time()
opt = lm.match_orbit_knobs_ip2_ip8(collider)
t4 = time.time()
optimizers['orbit_knobs'] = opt

# Rematch tunes and chromaticities
print('Rematching chromaticities')
for ll in collider.line_names:
    line = collider[ll]
    beam_name = ll[-2:]
    line.match(
        vary=[
            xt.Vary('kqtf.' + beam_name, step=1e-8),
            xt.Vary('kqtd.' + beam_name, step=1e-8),
            xt.Vary('ksf.' + beam_name, step=1e-8),
            xt.Vary('ksd.' + beam_name, step=1e-8),
        ],
        targets = [
            xt.Target('qx', tw0[ll].qx, tol=1e-6),
            xt.Target('qy', tw0[ll].qy, tol=1e-6),
            xt.Target('dqx', tw0[ll].dqx, tol=1e-4),
            xt.Target('dqy', tw0[ll].dqy, tol=1e-4)])

tw = collider.twiss()

for ll in collider.line_names:
    line = collider[ll]
    ddqx, ddqy = second_order_chromaticity(line)
    tw[ll]._data['ddqx'] = ddqx
    tw[ll]._data['ddqy'] = ddqy


# Tunes
print('Tunes:')
print(f"  b1: qx={tw.lhcb1.qx:6f} qy={tw.lhcb1.qy:6f}")
print(f"  b2: qx={tw.lhcb2.qx:6f} qy={tw.lhcb2.qy:6f}")

# Chromaticities
print('Chromaticities before:')
print(f"  b1: dqx={tw0.lhcb1.dqx:6f} dqy={tw0.lhcb1.dqy:6f}")
print(f"  b2: dqx={tw0.lhcb2.dqx:6f} dqy={tw0.lhcb2.dqy:6f}")
print('Chromaticities after:')
print(f"  b1: dqx={tw.lhcb1.dqx:6f} dqy={tw.lhcb1.dqy:6f}")
print(f"  b2: dqx={tw.lhcb2.dqx:6f} dqy={tw.lhcb2.dqy:6f}")

# Second order chromaticities
print('Second order chromaticities before:')
print(f"  b1: ddqx={tw0.lhcb1.ddqx:6f} ddqy={tw0.lhcb1.ddqy:6f}")
print(f"  b2: ddqx={tw0.lhcb2.ddqx:6f} ddqy={tw0.lhcb2.ddqy:6f}")
print('Second order chromaticities after:')
print(f"  b1: ddqx={tw.lhcb1.ddqx:6f} ddqy={tw.lhcb1.ddqy:6f}")
print(f"  b2: ddqx={tw.lhcb2.ddqx:6f} ddqy={tw.lhcb2.ddqy:6f}")

print(f'Time match IRs b1: {t_match_irs["b1"]}')
print(f'Time match IRs b2: {t_match_irs["b2"]}')
print(f'Time match orbit knobs: {t4-t3} s')

# # Open twisses to debug
# tw_b1_check = collider.lhcb1.twiss(
#     start='ip1', end='ip1.l1',
#     init=xt.TwissInit(betx=0.15, bety=0.15, element_name='ip1',
#                             line=collider.lhcb1))

# tw_b2_check = collider.lhcb2.twiss(
#     start='ip1', end='ip1.l1',
#     init=xt.TwissInit(betx=0.15, bety=0.15, element_name='ip1',
#                             line=collider.lhcb2))

# Save xsuite model
collider.to_json('collider_01_changed_ip15_phase.json')

# Save madx optics file
lm.gen_madx_optics_file_auto(collider, 'opt_round_150_1500_xs.madx')


import numpy as np
import xtrack as xt


# Load a line and build a tracker
# line = xt.Line.from_json('../../test_data/hllhc15_thick/lhc_thick_with_knobs.json')

line = xt.Line(elements=[xt.LineSegmentMap(qx=62.31, qy=60.32,
                        det_xx=1000, det_xy=10, det_yx=20, det_yy=2000)])
line.particle_ref = xt.Particles(p0c=7e9)
line.build_tracker()

nemitt_x = 2.5e-6
nemitt_y = 2.5e-6

det = line.get_amplitude_detuning_coefficients()

assert np.isclose(det['det_xx'], 1000, atol=1e-1, rtol=0)
assert np.isclose(det['det_yy'], 2000, atol=1e-1, rtol=0)
assert np.isclose(det['det_xy'], 10, atol=1e-1, rtol=0)
assert np.isclose(det['det_yx'], 20, atol=1e-1, rtol=0)




import xtrack as xt

env = xt.load_madx_lattice('../../test_data/psb_chicane/psb.seq')
env.vars.load_madx('../../test_data/psb_chicane/psb_fb_lhc.str')

line = env.psb1
line.particle_ref = xt.Particles(kinetic_energy0=2e9)

line.twiss4d().plot()


import xtrack as xt
import xdeps
import numpy as np
import re

from xdeps.madxutils import to_madx

stmt = re.compile(r"([a-z0-9._]+)(,|:=|:|=)(.*);")


def parse_command(stmt):
    groups = []
    rest = stmt
    while len(rest) > 0:
        if res := re.match(r"([a-z0-9._]+)[,;]", rest):
            groups.append(res.groups())
        elif res := re.match(r"([a-z0-9._]+)(=|:=)(\{[^\}]+\})[,;]", rest):
            groups.append(res.groups())
        elif res := re.match(r"([a-z0-9._]+)(=|:=)([^,]+)[,;]", rest):
            groups.append(res.groups())
        else:
            print(rest)
            raise ValueError("Input Error {stmt}")
        rest = rest[len(res.group()) :]
    return groups


# examples=["fd,b={4,6},a=5;"]
# for ex in examples:
#    print(parse_command(ex))


def parse_madx_expr(av):
    av_expr = env._xdeps_eval.eval(av)
    if not isinstance(av_expr, float):
        return av
    else:
        return av_expr


def invert_at(groups):
    out = []
    for an, eq, av in groups:
        if an == "at":
            av = f"lhclength-({av})"
        out.append((an, eq, av))
    return out


def groups_to_attrs(groups):
    attrs = {}
    extra = {}
    for an, eq, av in groups:
        if av.startswith("{"):
            av = [parse_madx_expr(aa) for aa in av[1:-1].split(",")]
        else:
            av = parse_madx_expr(av)
        if an in [
            "slot_id",
            "mech_sep",
            "assembly_id",
            "kmin",
            "kmax",
            "calib",
            "polarity",
        ]:
            extra[an] = (av, eq)
        else:
            if an == "l" or an == "lrad":
                an = "length"
            if an == "tilt":
                an = "rot_s_rad"
            attrs[an] = (av, eq)
    attrs = ",".join([f"{an}={av!r}" for an, (av, eq) in attrs.items()])
    extra = ",".join([f"{an}={av!r}" for an, (av, eq) in extra.items()])
    out = []
    if attrs:
        out.append(f"{attrs}")
    if extra:
        out.append(f"extra=dict({extra})")
    return "," + ",".join(out)


def default_for_zeros(expr, vv, notinitialized):
    if hasattr(expr, "_get_dependencies"):
        for ll in expr._get_dependencies():
            if hasattr(ll, "_key"):
                if ll._key not in vv:
                    vv[ll._key] = 0
                    notinitialized.add(ll._key)


env = xt.Environment()
env.vars["twopi"] = np.pi * 2

lines = [
    "import xtrack as xt",
    "import numpy as np",
    "env=xt.Environment()",
    "env._xdeps_vref._owner.default_factory = lambda : 0",
    "env.vars['twopi']=np.pi*2",
    "env.new('vkicker','Multipole')",
    "env.new('hkicker','Multipole')",
    "env.new('tkicker','Multipole')",
    "env.new('collimator','Drift')",
    "env.new('instrument','Drift')",
    "env.new('monitor','Drift')",
    "env.new('placeholder','Drift')",
    "env.new('sbend','Bend')",
    "env.new('rbend','Bend')",
    "env.new('quadrupole','Quadrupole')",
    "env.new('sextupole','Sextupole')",
    "env.new('octupole','Octupole')",
    "env.new('marker','Drift')",
    "env.new('rfcavity','Cavity')",
    "env.new('multipole','Multipole', knl=[0, 0, 0, 0, 0, 0])",
    "env.new('solenoid','Solenoid')",
]

notparsed = []
notinitialized = set()
inside_sequence = False
lhcb2 = False
for line in open("lhc.seq"):

    if '/*                       ACSCA CAVITIES ' in line:
        break

    line = line.replace(", L := l.OMK;", ";")
    line = line.replace(" L := l.ACSCA,", "")
    line = line.replace(", HARMON := HRF400", "")
    if "INSTRUMENT" in line or "PLACEHOLDER" in line:
        if "lrad" in line.lower():
            newline = line.lower()
            newline = newline.split("lrad")[0]
            newline = newline.strip()
            newline = newline.strip(",")
            newline = newline + ";"
            line = newline

    line = line.strip()
    ls = line.replace(" ", "").lower()
    if res := stmt.match(ls):
        lhs, eq, rhs = res.groups()
        if eq == ":=":  # variable
            expr = env._xdeps_eval.eval(rhs)
            default_for_zeros(expr, env.vars, notinitialized)
            env.vars[lhs] = expr
            if not isinstance(expr, float):
                expr = rhs
            lines.append(f"env[{lhs!r}]={expr!r}")
        elif eq == ":":  # element definition
            groups = parse_command(rhs + ";")
            parent = groups.pop(0)[0]
            if parent == "sequence":
                lines.append(f"{lhs}=env.new_builder(name={lhs!r})")
                inside_sequence = True
                sequence_name=lhs
                if lhs == "lhcb2":
                    lhcb2 = True
            else:
                groups = [
                    (an.replace("from", "from_"), eq, av) for an, eq, av in groups
                ]
                if lhcb2:
                    groups = invert_at(groups)
                attrs = groups_to_attrs(groups)
                if inside_sequence:
                    lines.append(f"{sequence_name}.new({lhs!r:20},{parent!r}{attrs})")
                else:
                    lines.append(f"env.new({lhs!r:20},{parent!r}{attrs})")
        elif eq == ",":  # element definition
            groups = parse_command(rhs + ";")
            attrs = groups_to_attrs(groups)
            lines.append(f"env.set({lhs!r:20}{attrs})")
        else:
            lines.append("#" + line)
    else:
        if line.startswith("ENDSEQUENCE"):
            if lhcb2:
               lines.append(f"#{sequence_name}.reflect().build() ")
            else:
               lines.append(f"{sequence_name}.build()")
            inside_sequence = False
        elif len(line) > 0:
            if not line.startswith("if") or not line.startswith("return"):
                lines.append("#" + line)
        else:
            lines.append("")

import re

for ii in range(len(lines)):
    line = lines[ii]
    if ",kick" in line:
        if "h." in line:
            line = re.sub(r"kick='([-a-z0-9.]+)'", r"knl=['-(\1)']", line)
        else:
            line = re.sub(r"kick=('[-a-z0-9.]+')", r"ksl=[\1]", line)
    if ",hkick" in line:
        line = re.sub(r"hkick='([-a-z0-9.]+)'", r"knl=['-(\1)']", line)
    if ",vkick" in line:
        line = re.sub(r"vkick=('[-a-z0-9./]+')", r"ksl=[\1]", line)
    lines[ii] = line

for ll in lines:
    print(ll)



import pathlib

import xobjects as xo
import xtrack as xt
from xobjects.test_helpers import for_all_test_contexts

test_data_folder = pathlib.Path(
    __file__).parent.joinpath('../test_data').absolute()

@for_all_test_contexts
def test_cavity_absolute_time(test_context):
    line = xt.Line.from_json(test_data_folder /
                            'hllhc15_thick/lhc_thick_with_knobs.json')
    line.build_tracker(_context=test_context)
    line.vv['vrf400'] = 16

    for vv in line.vars.get_table().rows[
        'on_x.*|on_sep.*|on_crab.*|on_alice|on_lhcb|corr_.*'].name:
        line.vars[vv] = 0

    tw = line.twiss()

    df_hz = -50
    h_rf = 35640
    f_rev = 1/tw.T_rev0
    df_rev = df_hz / h_rf
    eta = tw.slip_factor

    f_rf0 = 1/tw.T_rev0 * h_rf

    f_rf = f_rf0 + df_hz
    line.vars['f_rf'] = f_rf
    tt = line.get_table()
    for nn in tt.rows[tt.element_type=='Cavity'].name:
        line.element_refs[nn].absolute_time = 1 # Need property
        line.element_refs[nn].frequency = line.vars['f_rf']

    tw1 = line.twiss(search_for_t_rev=True)

    f_rev_expected = f_rf / h_rf

    xo.assert_allclose(f_rev_expected, 1/tw1.T_rev, atol=1e-5, rtol=0)
    xo.assert_allclose(tw1.delta, tw1.delta[0], atol=1e-5, rtol=0) # Check that it is flat
    delta_expected = -df_rev / f_rev / eta
    xo.assert_allclose(tw1.delta, delta_expected, atol=2e-6, rtol=0)
    tw_off_mom = line.twiss(method='4d', delta0=tw1.delta[0])
    xo.assert_allclose(tw1.x, tw_off_mom.x, atol=1e-5, rtol=0)


import numpy as np
from cpymad.madx import Madx
import xtrack as xt

# Import a line and build a tracker
line = xt.Line.from_json(
    '../../test_data/psb_injection/line_and_particle.json')
e_kin_start_eV = 160e6
line.particle_ref = xt.Particles(mass0=xt.PROTON_MASS_EV, q0=1.,
                                 energy0=xt.PROTON_MASS_EV + e_kin_start_eV)
line.build_tracker()

tw0 = line.twiss4d()

# User-defined energy ramp
t_s = np.array([0., 0.0006, 0.0008, 0.001 , 0.0012, 0.0014, 0.0016, 0.0018,
                0.002 , 0.0022, 0.0024, 0.0026, 0.0028, 0.003, 0.01, 0.1])
E_kin_GeV = np.array([0.16000000,0.16000000,
    0.16000437, 0.16001673, 0.16003748, 0.16006596, 0.16010243, 0.16014637,
    0.16019791, 0.16025666, 0.16032262, 0.16039552, 0.16047524, 0.16056165,
    0.163586, 0.20247050000000014])

# Attach energy program to the line
line.energy_program = xt.EnergyProgram(t_s=t_s, kinetic_energy0=E_kin_GeV*1e9)

# Plot energy and revolution frequency vs time
t_plot = np.linspace(0, 10e-3, 20)
E_kin_plot = line.energy_program.get_kinetic_energy0_at_t_s(t_plot)
f_rev_plot = line.energy_program.get_frev_at_t_s(t_plot)

import matplotlib.pyplot as plt
plt.close('all')
plt.figure(1, figsize=(6.4 * 1.5, 4.8))
ax1 = plt.subplot(2,2,1)
plt.plot(t_plot * 1e3, E_kin_plot * 1e-6)
plt.ylabel(r'$E_{kin}$ [MeV]')
ax2 = plt.subplot(2,2,3, sharex=ax1)
plt.plot(t_plot * 1e3, f_rev_plot * 1e-3)
plt.ylabel(r'$f_{rev}$ [kHz]')
plt.xlabel('t [ms]')

# Setup frequency of the RF cavity  to stay on the second harmonic of the
# revolution frequency during the acceleration

t_rf = np.linspace(0, 3e-3, 100) # time samples for the frequency program
f_rev = line.energy_program.get_frev_at_t_s(t_rf)
h_rf = 2 # harmonic number
f_rf = h_rf * f_rev # frequency program

# Build a function with these samples and link it to the cavity
line.functions['fun_f_rf'] = xt.FunctionPieceWiseLinear(x=t_rf, y=f_rf)
line.element_refs['br.c02'].frequency = line.functions['fun_f_rf'](
                                                        line.vars['t_turn_s'])

# Setup voltage and lag
line.element_refs['br.c02'].voltage = 3000 # V
line.element_refs['br.c02'].lag = 0 # degrees (below transition energy)

# When setting line.vars['t_turn_s'] the reference energy and the rf frequency
# are updated automatically
line.vars['t_turn_s'] = 0
line.particle_ref.kinetic_energy0 # is 160.00000 MeV
line['br.c02'].frequency # is 1983931.935 Hz

line.vars['t_turn_s'] = 3e-3
line.particle_ref.kinetic_energy0 # is 160.56165 MeV
line['br.c02'].frequency # is 1986669.0559674294

# Back to zero for tracking!
line.vars['t_turn_s'] = 0

# Track a few particles to visualize the longitudinal phase space
p_test = line.build_particles(x_norm=0, zeta=np.linspace(0, line.get_length(), 101))

# Enable time-dependent variables (t_turn_s and all the variables that depend on
# it are automatically updated at each turn)
line.enable_time_dependent_vars = True

# Track
line.track(p_test, num_turns=9000, turn_by_turn_monitor=True, with_progress=True)
mon = line.record_last_track

# Plot
plt.subplot2grid((2,2), (0,1), rowspan=2)
plt.plot(mon.zeta[:, -2000:].T, mon.delta[:, -2000:].T, color='C0')
plt.xlabel(r'$\zeta$ [m]')
plt.ylabel('$\delta$')
plt.xlim(-40, 30)
plt.ylim(-0.0025, 0.0025)
plt.title('Last 2000 turns')
plt.subplots_adjust(left=0.08, right=0.95, wspace=0.26)


# Check transverse beam size reduction
line['t_turn_s'] = 0
line.enable_time_dependent_vars = False

n_part_test = 500
# Generate Gaussian distribution with fixed rng seed
rng = np.random.default_rng(seed=123)
x_norm = rng.normal(loc=0, scale=1, size=n_part_test)
px_norm = rng.normal(loc=0, scale=1, size=n_part_test)
y_norm = rng.normal(loc=0, scale=1, size=n_part_test)
py_norm = rng.normal(loc=0, scale=1, size=n_part_test)

# rescale to have exact std dev.
x_norm = x_norm / np.std(x_norm)
px_norm = px_norm / np.std(px_norm)
y_norm = y_norm / np.std(y_norm)
py_norm = py_norm / np.std(py_norm)

p_test2 = line.build_particles(x_norm=x_norm, px_norm=px_norm,
                               y_norm=x_norm, py_norm=px_norm,
                               nemitt_x=3e-6, nemitt_y=3e-6,
                               delta=0)

line.enable_time_dependent_vars = True
line.track(p_test2, num_turns=50_000, turn_by_turn_monitor=True, with_progress=True)
mon2 = line.record_last_track

std_y = np.std(mon2.y, axis=0)
std_x = np.std(mon2.x, axis=0)

# Apply moving average filter
from scipy.signal import savgol_filter
std_y_smooth = savgol_filter(std_y, 10000, 2)
std_x_smooth = savgol_filter(std_x, 10000, 2)

i_turn_match = 10000
std_y_expected = std_y_smooth[i_turn_match] * np.sqrt(
    mon2.gamma0[0, i_turn_match]* mon2.beta0[0, i_turn_match]
    / mon2.gamma0[0, :] / mon2.beta0[0, :])
std_x_expected = std_x_smooth[i_turn_match] * np.sqrt(
    mon2.gamma0[0, i_turn_match]* mon2.beta0[0, i_turn_match]
    / mon2.gamma0[0, :] / mon2.beta0[0, :])

d_sigma_x = std_x_expected[0] - std_x_expected[-1]
d_sigma_y = std_y_expected[0] - std_y_expected[-1]

import xobjects as xo
xo.assert_allclose(std_y_expected[40000:45000].mean(),
                   std_y_smooth[40000:45000].mean(),
                   rtol=0, atol=0.07 * d_sigma_y)
xo.assert_allclose(std_x_expected[40000:45000].mean(),
                   std_x_smooth[40000:45000].mean(),
                   rtol=0, atol=0.07 * d_sigma_x)

plt.figure(2)
ax1 = plt.subplot(2,1,1)
plt.plot(std_x, label='raw')
plt.plot(std_x_smooth, label='smooth')
plt.plot(std_x_expected, label='expected')
plt.legend()

ax2 = plt.subplot(2,1,2, sharex=ax1)
plt.plot(std_y, label='raw')
plt.plot(std_y_smooth, label='smooth')
plt.plot(std_y_expected, label='expected')


plt.show()



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import json
import numpy as np

import xobjects as xo
import xtrack as xt

###############
# Load a line #
###############

fname_line_particles = '../../test_data/hllhc15_noerrors_nobb/line_and_particle.json'
#fname_line_particles = '../../test_data/sps_w_spacecharge/line_no_spacecharge_and_particle.json' #!skip-doc

with open(fname_line_particles, 'r') as fid:
    input_data = json.load(fid)
line = xt.Line.from_dict(input_data['line'])
line.particle_ref = xt.Particles.from_dict(input_data['particle'])

line.build_tracker()

starting = {
    "theta0": -np.pi / 9,
    "psi0": np.pi / 7,
    "phi0": np.pi / 11,
    "X0": -300,
    "Y0": 150,
    "Z0": -100,
}


line_c = line.cycle('ip5')

for reverse in [False, True]:

    sv0 = line.survey(element0='ip5', **starting)
    sv_c = line_c.survey(**starting)
    if reverse:
        sv0=sv0.reverse()
        sv_c=sv_c.reverse()

    sv0 = sv0.to_pandas()
    sv_c = sv_c.to_pandas()

    for ename in ['ip5', 'ip8', 'ip1', 'mb.c12r8.b1..1',
                'mb.c12r1.b1..1', 'mb.c12l5.b1..1', 'drift_10', 'drift_10000']:
        svc_at_e = sv_c[sv_c.name==ename]
        sv0_at_e = sv0[sv0.name==ename]

        assert np.isclose(svc_at_e.X, sv0_at_e.X, rtol=0, atol=5e-4)
        assert np.isclose(svc_at_e.Y, sv0_at_e.Y, rtol=0, atol=5e-4)
        assert np.isclose(svc_at_e.Z, sv0_at_e.Z, rtol=0, atol=5e-4)
        assert np.isclose(np.mod(svc_at_e.theta, 2*np.pi), np.mod(sv0_at_e.theta, 2*np.pi),
                        rtol=0, atol=1e-14)
        assert np.isclose(svc_at_e.phi, sv0_at_e.phi, rtol=0, atol=5e-14)
        assert np.isclose(svc_at_e.psi, sv0_at_e.psi, rtol=0, atol=5e-14)
        assert np.isclose(svc_at_e.drift_length, sv0_at_e.drift_length, rtol=0, atol=1e-14)
        assert np.isclose(svc_at_e.angle, sv0_at_e.angle, rtol=0, atol=1e-14)
        assert np.isclose(svc_at_e.tilt, sv0_at_e.tilt, rtol=0, atol=1e-14)



import xtrack as xt
import xobjects as xo
import numpy as np
from scipy.constants import c as clight

p = xt.Particles(mass0=xt.PROTON_MASS_EV, q0=1,
                 kinetic_energy0=50e6, delta=0.1)

beta0 = p.beta0[0]
gamma0 = p.gamma0[0]
mass0_ev = p.mass0
energy0 = p.energy0[0]
energy = p.energy[0]
p0c = p.p0c[0]

delta = p.delta[0]
ptau = p.ptau[0]
pzeta = p.pzeta[0]

beta = beta0 * p.rvv[0]
gamma = energy / energy0 * gamma0

xo.assert_allclose(gamma0, 1 / np.sqrt(1 - beta0**2), rtol=0, atol=1e-14)
xo.assert_allclose(gamma, 1 / np.sqrt(1 - beta**2), rtol=0, atol=1e-14)
xo.assert_allclose(energy, mass0_ev * gamma, rtol=0, atol=1e-6) # 1e-6 eV
xo.assert_allclose(energy0, mass0_ev * gamma0, rtol=0, atol=1e-6) # 1e-6 eV

Pc = p0c * (1 + delta)
xo.assert_allclose(Pc, mass0_ev * gamma * beta, rtol=0, atol=1e-6) # 1e-6 eV

# Definitions delta/ptau/pzeta
xo.assert_allclose(delta, (Pc - p0c) / p0c, rtol=0, atol=1e-14)
xo.assert_allclose(ptau, (energy - energy0) / energy0 / beta0, rtol=0, atol=1e-14)
xo.assert_allclose(pzeta,(energy - energy0) / energy0 / beta0**2, rtol=0, atol=1e-14)
xo.assert_allclose(pzeta,(energy - energy0) / p0c / beta0, rtol=0, atol=1e-14)

# Conversions
xo.assert_allclose(delta, np.sqrt(ptau**2 + 2*ptau/beta0 + 1) - 1, rtol=0, atol=1e-14)
xo.assert_allclose(delta, np.sqrt(beta0**2 * pzeta**2 + 2*pzeta + 1) - 1, rtol=0, atol=1e-14)
xo.assert_allclose(delta, beta * ptau + (beta - beta0)/beta0, rtol=0, atol=1e-14)
xo.assert_allclose(delta, beta * beta0 * pzeta + (beta - beta0)/beta0, rtol=0, atol=1e-14)

# More relations from the physics manual
xo.assert_allclose(gamma, gamma0 * (1 + beta0 * ptau), rtol=0, atol=1e-14)
xo.assert_allclose(beta, np.sqrt(1 - (1 - beta0**2)/(1 + beta0 * ptau)**2), rtol=0, atol=1e-14)

# Check relations for small energy deviations
p_small = xt.Particles(mass0=xt.PROTON_MASS_EV, q0=1,
                       kinetic_energy0=50e6, delta=1e-4)
delta_small = p_small.delta[0]
ptau_small = p_small.ptau[0]
pzeta_small = p_small.pzeta[0]
beta_small = beta0 * p_small.rvv[0]

xo.assert_allclose(delta_small, ptau_small/beta0, rtol=0, atol=1e-8)
xo.assert_allclose(delta_small, pzeta_small, rtol=0, atol=1e-8)
xo.assert_allclose(beta_small, beta0 + (1 - beta0**2) * ptau_small, rtol=0, atol=1e-8)


# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

import numpy as np
from numbers import Number

from .masses import U_MASS_EV
from .masses import __dict__ as mass__dict__

# Monte Carlo numbering scheme as defined by the Particle Data Group
# See https://pdg.lbl.gov/2007/reviews/montecarlorpp.pdf for implementation
# details.
# Not all particles are implemented yet; this can be appended later


# This is the internal dictionary of PDG IDs.
# For each pdg id, we get the charge and a list of names.
# The first name is the default short name, the last is the default long name.
# The other names are accepted alternatives. Any superscript or subscript can be used
# interchangeably with its normal script (if both are used, subscript comes first).

pdg_table = {
#   ID       q  NAME
    0:     [0,  'undefined'],
    11:    [-1, 'e⁻', 'e', 'electron'],
    -11:   [1,  'e⁺', 'positron'],
    12:    [0,  'νₑ', 'electron neutrino'],
    13:    [-1, 'μ⁻', 'μ', 'muon-', 'muon'],
    -13:   [1,  'μ⁺', 'muon+', 'anti-muon'],
    14:    [0,  'νμ', 'muon neutrino'],
    15:    [-1, 'τ⁻', 'τ', 'tau-', 'tau'],
    -15:   [1, 'τ⁺', 'tau+', 'anti-tau'],
    16:    [0,  'ντ', 'tau neutrino'],
    22:    [0,  'γ⁰', 'γ', 'gamma', 'photon'],
    111:   [0,  'π⁰', 'π', 'pion', 'pion0', 'pi0'],
    211:   [1,  'π⁺', 'pion+', 'pi+'],
    -211:  [-1, 'π⁻', 'pion-', 'pi-'],
    311:   [0,  'K⁰', 'kaon', 'kaon0'],
    321:   [1,  'K⁺', 'kaon+'],
    -321:  [-1, 'K⁻', 'kaon-'],
    130:   [0,  'KL', 'long kaon'],
    310:   [0,  'Kₛ', 'short kaon'],
    421:   [0,  'D⁰', 'D'],
    411:   [1,  'D⁺'],
    -411:  [-1, 'D⁻'],
    431:   [1,  'Dₛ⁺'],
    -431:  [-1, 'Dₛ⁻'],
    2212:  [1,  'p⁺', 'p', 'proton'],
    -2212: [-1,  'p⁻', 'anti-proton'],
    2112:  [0,  'n⁰', 'n', 'neutron'],
    2224:  [2,  'Δ⁺⁺', 'delta++'],
    2214:  [1,  'Δ⁺', 'delta+'],
    2114:  [0,  'Δ⁰', 'delta0'],
    1114:  [-1, 'Δ⁻', 'delta-'],
    3122:  [0,  'Λ⁰', 'Λ', 'lambda'],
    4122:  [1,  'Λc⁺', 'lambdac+'],
    3222:  [1,  'Σ⁺', 'sigma+'],
    3212:  [0,  'Σ⁰', 'Σ', 'sigma', 'sigma0'],
    3112:  [-1, 'Σ⁻', 'sigma-'],
    3322:  [0,  'Ξ⁰', 'Ξ', 'xi', 'xi0'],
    3312:  [-1, 'Ξ⁻', 'xi-'],
    4132:  [0,  'Ξc⁰', 'Ξc', 'xic', 'xic0'],
    4232:  [1,  'Ξc⁺', 'xic+'],
    4312:  [0,  "Ξ'c⁰", "Ξ'c", "xiprimec", "xiprimec0"],
    4322:  [1,  "Ξ'c⁺", "xiprimec+"],
    3334:  [-1, 'Ω⁻', 'omega-'],
    4332:  [0, 'Ωc⁰', 'Ωc', 'omegac', 'omegac0'],
    1000010020: [1, '²H', 'H2', 'hydrogen-2', 'deuteron'],
    1000010030: [1, '³H', 'H3', 'hydrogen-3', 'triton']
}

elements = {
     1: "H",    2: "He",   3: "Li",   4: "Be",   5: "B",    6: "C",    7: "N",    8: "O",    9: "F",   10: "Ne",
    11: "Na",  12: "Mg",  13: "Al",  14: "Si",  15: "P",   16: "S",   17: "Cl",  18: "Ar",  19: "K",   20: "Ca",
    21: "Sc",  22: "Ti",  23: "V",   24: "Cr",  25: "Mn",  26: "Fe",  27: "Co",  28: "Ni",  29: "Cu",  30: "Zn",
    31: "Ga",  32: "Ge",  33: "As",  34: "Se",  35: "Br",  36: "Kr",  37: "Rb",  38: "Sr",  39: "Y",   40: "Zr",
    41: "Nb",  42: "Mo",  43: "Tc",  44: "Ru",  45: "Rh",  46: "Pd",  47: "Ag",  48: "Cd",  49: "In",  50: "Sn",
    51: "Sb",  52: "Te",  53: "I",   54: "Xe",  55: "Cs",  56: "Ba",  57: "La",  58: "Ce",  59: "Pr",  60: "Nd",
    61: "Pm",  62: "Sm",  63: "Eu",  64: "Gd",  65: "Tb",  66: "Dy",  67: "Ho",  68: "Er",  69: "Tm",  70: "Yb",
    71: "Lu",  72: "Hf",  73: "Ta",  74: "W",   75: "Re",  76: "Os",  77: "Ir",  78: "Pt",  79: "Au",  80: "Hg",
    81: "Tl",  82: "Pb",  83: "Bi",  84: "Po",  85: "At",  86: "Rn",  87: "Fr",  88: "Ra",  89: "Ac",  90: "Th",
    91: "Pa",  92: "U",   93: "Np",  94: "Pu",  95: "Am",  96: "Cm",  97: "Bk",  98: "Cf",  99: "Es", 100: "Fm",
   101: "Md", 102: "No", 103: "Lr", 104: "Rf", 105: "Db", 106: "Sg", 107: "Bh", 108: "Hs", 109: "Mt", 110: "Ds",
   111: "Rg", 112: "Cn", 113: "Nh", 114: "Fl", 115: "Mc", 116: "Lv", 117: "Ts", 118: "Og"
}
_elements_inv = {vv: kk for kk, vv in elements.items()}

elements_long = {
    1: "Hydrogen",        2: "Helium",          3: "Lithium",         4: "Beryllium",       5: "Boron",
    6: "Carbon",          7: "Nitrogen",        8: "Oxygen",          9: "Fluorine",       10: "Neon",
   11: "Sodium",         12: "Magnesium",      13: "Aluminum",       14: "Silicon",        15: "Phosphorus",
   16: "Sulfur",         17: "Chlorine",       18: "Argon",          19: "Potassium",      20: "Calcium",
   21: "Scandium",       22: "Titanium",       23: "Vanadium",       24: "Chromium",       25: "Manganese",
   26: "Iron",           27: "Cobalt",         28: "Nickel",         29: "Copper",         30: "Zinc",
   31: "Gallium",        32: "Germanium",      33: "Arsenic",        34: "Selenium",       35: "Bromine",
   36: "Krypton",        37: "Rubidium",       38: "Strontium",      39: "Yttrium",        40: "Zirconium",
   41: "Niobium",        42: "Molybdenum",     43: "Technetium",     44: "Ruthenium",      45: "Rhodium",
   46: "Palladium",      47: "Silver",         48: "Cadmium",        49: "Indium",         50: "Tin",
   51: "Antimony",       52: "Tellurium",      53: "Iodine",         54: "Xenon",          55: "Cesium",
   56: "Barium",         57: "Lanthanum",      58: "Cerium",         59: "Praseodymium",   60: "Neodymium",
   61: "Promethium",     62: "Samarium",       63: "Europium",       64: "Gadolinium",     65: "Terbium",
   66: "Dysprosium",     67: "Holmium",        68: "Erbium",         69: "Thulium",        70: "Ytterbium",
   71: "Lutetium",       72: "Hafnium",        73: "Tantalum",       74: "Tungsten",       75: "Rhenium",
   76: "Osmium",         77: "Iridium",        78: "Platinum",       79: "Gold",           80: "Mercury",
   81: "Thallium",       82: "Lead",           83: "Bismuth",        84: "Polonium",       85: "Astatine",
   86: "Radon",          87: "Francium",       88: "Radium",         89: "Actinium",       90: "Thorium",
   91: "Protactinium",   92: "Uranium",        93: "Neptunium",      94: "Plutonium",      95: "Americium",
   96: "Curium",         97: "Berkelium",      98: "Californium",    99: "Einsteinium",   100: "Fermium",
  101: "Mendelevium",   102: "Nobelium",      103: "Lawrencium",    104: "Rutherfordium", 105: "Dubnium",
  106: "Seaborgium",    107: "Bohrium",       108: "Hassium",       109: "Meitnerium",    110: "Darmstadtium",
  111: "Roentgenium",   112: "Copernicium",   113: "Nihonium",      114: "Flerovium",     115: "Moscovium",
  116: "Livermorium",   117: "Tennessine",    118: "Oganesson"
}
_elements_long_inv = {vv: kk for kk, vv in elements_long.items()}


def is_proton(pdg_id):
    """Check if a PDG ID corresponds to a proton."""
    return int(pdg_id) == 2212

def is_ion(pdg_id):
    """Check if a PDG ID corresponds to a heavy ion (A+Z > 1)."""
    tmpid = pdg_id - 1000000000
    L = int(tmpid/1e7)
    tmpid -= L*1e7
    Z = int(tmpid /1e4)
    tmpid -= Z*1e4
    A = int(tmpid /10)
    return (int(pdg_id) >= 1000000000) and (Z > 0) and (A > Z)

def is_lepton(pdg_id):
    """Check if a PDG ID corresponds to a lepton (neutrinos included)."""
    return 11 <= abs(int(pdg_id)) <= 16


def get_name_from_pdg_id(pdg_id, long_name=True, subscripts=True):
    """
    Get the name of a particle from its PDG ID.

    Parameters
    ----------
    long_name : bool, default True
        If True, return the long name of the particle (ASCII-compliant). For a
        particle in the PDF internal table, this is the last name in the list
        of alternatives. For an ion, it is the full element name, followed by
        the mass number. If 'long_name' is False, a short name is returned
        (Unicode). For a particle in the PDF internal table, this is the first
        name in the list of alternatives. For an ion, it is the element name
        preceded (when 'subscripts' is True) of followed (when 'subscripts' is
        False) by the mass number.
    subscripts : bool, default True
        Controls whether or not to allow sub- and superscripts in the short
        name. Has no function if 'long_name' is True.

    Returns
    -------
    str
        The name of the particle.
    """
    if hasattr(pdg_id, '__len__') and not isinstance(pdg_id, str):
        return np.array([get_name_from_pdg_id(pdg) for pdg in pdg_id])
    return get_properties_from_pdg_id(pdg_id, long_name=long_name, subscripts=subscripts)[-1]


def get_pdg_id_from_name(name=None):
    """
    Get a particle's PDG ID from its name.

    Parameters
    ----------
    name : str
        The name of the particle. Can be any alternative from the PDF internal
        table, with or without sub- or superscripts. For ions, the name can be
        any combination of the element short or long name and the mass number,
        for instance 'Pb208', 'Pb 208', Pb-208', 'Pb_208', '208Pb', 'Lead-208',
        'lead 208', etc.

    Returns
    -------
    int
        The PDG ID of the particle.
    """
    if hasattr(name, 'get'):
        name = name.get()

    if name is None:
        return 0  # undefined
    elif hasattr(name, '__len__') and not isinstance(name, str):
        return np.array([get_pdg_id_from_name(nn) for nn in name])
    elif isinstance(name, Number):
        return int(name) # fallback

    lname = _to_normal_script(name).lower().replace('ς', 'σ')
    aname = ""
    if len(lname) > 4 and lname[:5]=="anti-":
        aname = _flip_end_sign(lname[5:])
    elif len(lname) > 4 and lname[:5]=="anti ":
        aname = _flip_end_sign(lname[5:])
    elif len(lname) > 3 and lname[:4]=="anti":
        aname = _flip_end_sign(lname[4:])

    _PDG_inv = get_pdg_inv()

    # particle
    if lname in _PDG_inv.keys():
        return _PDG_inv[lname]

    # anti-particle
    elif aname in _PDG_inv.keys():
        return -_PDG_inv[aname]

    else:
        ion_name = lname.lower()
        ion_name = ion_name.replace('_','').replace('-','')
        ion_name = ion_name.replace(' ','').replace('.','')
        for Z, ion in elements_long.items():
            if ion.lower() in ion_name:
                A = ion_name.replace(ion.lower(), '')
                if A.isnumeric() and int(A) > 0:
                    return get_pdg_id_ion(int(A), Z)
        for Z, ion in elements.items():
            if ion.lower() in ion_name:
                A = ion_name.replace(ion.lower(), '')
                if A.isnumeric() and int(A) > 0:
                    return get_pdg_id_ion(int(A), Z)
        raise ValueError(f"Particle {name} not found in pdg dictionary, or wrongly "
                        + f"formatted ion name!\nFor ions, use e.g. 'Pb208', 'Pb 208', "
                        + f"'Pb-208', 'Pb_208', 'Pb.208', '208Pb', 'lead-208', ...")


def get_properties_from_pdg_id(pdg_id, long_name=False, subscripts=True):
    """
    Get the properties of a particle from its PDG ID.

    Parameters
    ----------
    pdg_id : int or str
        The PDG ID of the particle.
    long_name : bool, default True
        If True, return the long name of the particle (ASCII-compliant). For a
        particle in the PDF internal table, this is the last name in the list
        of alternatives. For an ion, it is the full element name, followed by
        the mass number. If 'long_name' is False, a short name is returned
        (Unicode). For a particle in the PDF internal table, this is the first
        name in the list of alternatives. For an ion, it is the element name
        preceded (when 'subscripts' is True) of followed (when 'subscripts' is
        False) by the mass number.
    subscripts : bool, default True
        Controls whether or not to allow sub- and superscripts in the short
        name. Has no function if 'long_name' is True.
    name : str

    Returns
    -------
    q : int
        The charge of the particle.
    A : int
        The mass number of the particle (total number of protons and neutrons).
    Z : int
        The atomic number of the particle (number of protons).
    name : str
        The name of the particle, with formatting depending on 'long_name' and
        'subscripts'.
    """
    if hasattr(pdg_id, '__len__') and not isinstance(pdg_id, str):
        result = np.array([get_properties_from_pdg_id(pdg) for pdg in pdg_id]).T
        return result[0].astype(np.float64), result[1].astype(np.int64),\
               result[2].astype(np.int64), result[3]

    pdg_id = int(pdg_id)

    if pdg_id in pdg_table.keys():
        q = pdg_table[pdg_id][0]
        if long_name:
            name = pdg_table[pdg_id][-1]
        else:
            name = pdg_table[pdg_id][1]
            if not subscripts:
                name = _to_normal_script(name)
        if abs(pdg_id)==2212 or abs(pdg_id)==2112:
            A = 1
            Z = q
        elif pdg_id==1000010020:
            A = 2
            Z = 1
            if not long_name and not subscripts:
                name = 'H2'
        elif pdg_id==1000010030:
            A = 3
            Z = 1
            if not long_name and not subscripts:
                name = 'H3'
        else:
            A = 0
            Z = 0
        return int(q), int(A), int(Z), name

    elif pdg_id > 1000000000:
        # Ion
        tmpid = pdg_id - 1000000000
        L = int(tmpid/1e7)
        tmpid -= L*1e7
        Z = int(tmpid /1e4)
        tmpid -= Z*1e4
        A = int(tmpid /10)
        tmpid -= A*10
        isomer_level = int(tmpid)
        if long_name:
            return int(Z), int(A), int(Z), f'{get_element_full_name_from_Z(Z)}-{A}'
        else:
            if subscripts:
                name = f'{_digits_to_superscript(A)}{get_element_name_from_Z(Z)}'
            else:
                name = f'{get_element_name_from_Z(Z)}{A}'
            return int(Z), int(A), int(Z), name

    elif -pdg_id in pdg_table.keys() or pdg_id < -1000000000:
        antipart = get_properties_from_pdg_id(-pdg_id, long_name=long_name, subscripts=subscripts)
        name = _flip_end_sign(f'anti-{antipart[3]}')
        return -antipart[0], antipart[1], -antipart[2], name

    else:
        raise ValueError(f"PDG ID {pdg_id} not recognised!")


def get_element_name_from_Z(z):
    """Get the short element name for the element with Z protons."""
    if z not in elements:
        raise ValueError(f"Element with {z} protons not known.")
    return elements[z]

def get_element_full_name_from_Z(z):
    """Get the full element name for the element with Z protons."""
    if z not in elements_long:
        raise ValueError(f"Element with {z} protons not known.")
    return elements_long[z]

def get_Z_from_element_name(name):
    """Get the atomic number from the short or long element name."""
    if name in _elements_inv:
        return _elements_inv[name]
    elif name in _elements_long_inv:
        return _elements_long_inv[name]
    else:
        raise ValueError(f"Element {name} not recognised!")


def get_pdg_id_ion(A, Z):
    """Get the PDG ID for an ion with Z protons and A-Z neutrons."""
    if hasattr(A, '__len__') and not isinstance(A, str) \
    and hasattr(Z, '__len__') and not isinstance(Z, str):
        return np.array([get_pdg_id_ion(aa, zz) for aa, zz in zip(A,Z)])
    Z = int(Z)
    A = int(A)
    if Z < 1 or A < 2 or A < Z:
        raise ValueError(f"Not an ion ({Z=} and {A=})!")
    return 1000000000 + Z*10000 + A*10


def get_pdg_id_from_mass_charge(m, q, rtol=1e-3):
    """Get the PDG ID for a given mass and charge. This is always an estimate."""
    if hasattr(q, '__len__') and not isinstance(q, str) \
    and hasattr(m, '__len__') and not isinstance(m, str):
        return np.array([get_pdg_id_from_mass_charge(mm, qq) for qq, mm in zip(q, m)])
    elif hasattr(q, '__len__') and not isinstance(q, str):
        return np.array([get_pdg_id_from_mass_charge(m, qq) for qq in q])
    elif hasattr(m, '__len__') and not isinstance(m, str):
        return np.array([get_pdg_id_from_mass_charge(mm, q) for mm in m])

    m = float(m)
    q = int(q)
    A = round(m/U_MASS_EV)

    # First we check the internal table of masses
    found_ids = []
    for pdg_id, val in _get_mass_table().items():
        if abs(val) < 1e-12:
            if abs(m) < 1e-12:
                found_ids.append(pdg_id)
        elif abs(val-m)/val <= rtol:
            found_ids.append(pdg_id)
    if len(found_ids) > 0:
        pdg_id = []
        for this_pdg_id in found_ids:
            this_q, _, _, _ = get_properties_from_pdg_id(this_pdg_id)
            if this_q == q:
                pdg_id.append(this_pdg_id)
            else:
                this_q, _, _, _ = get_properties_from_pdg_id(-this_pdg_id)
                if this_q == q:
                    pdg_id.append(-this_pdg_id)
        if len(pdg_id) == 1:
            return pdg_id[0]
        elif len(pdg_id) > 1:
            raise ValueError(f"Multiple particles found for {m=} and {q=}: {found_ids}. "
                           + f"Decrease the tolerance. If this does not work, it might be "
                           + f"ambiguous and not resolvable with mass and charge alone "
                           + f"(like for a neutral particle that is not its own anti-particle).")

    # Then we check for ions
    if q > 0 and A > 1:
        return get_pdg_id_ion(A, q)

    # Not found
    else:
        raise ValueError(f"Particle with charge {q} and mass {m} eV not recognised!\nIf "
                       + f"the particle mass is expected to be present in the internal "
                       + f"mass table in Xtrack, try increasing the tolerance.")


def get_mass_from_pdg_id(pdg_id, allow_approximation=True, expected_mass=None, verbose=True):
    """Get the particle mass for a given PDG ID, if in the internal database. If not, it can be extrapolated for ions."""
    if hasattr(pdg_id, '__len__') and not isinstance(pdg_id, str):
        return np.array([get_mass_from_pdg_id(pdg,
                                      allow_approximation=allow_approximation,
                                      expected_mass=expected_mass)
                         for pdg in pdg_id])

    pdg_id = int(pdg_id)
    _mass_table = _get_mass_table()
    if pdg_id in _mass_table:
        return _mass_table[pdg_id]
    elif -pdg_id in _mass_table:
        return _mass_table[-pdg_id]
    elif pdg_id > 1000000000 and allow_approximation:
        _, A, _, _ = get_properties_from_pdg_id(pdg_id)
        if verbose:
            print(f"Warning: approximating the mass as {A}u!")
        return A*U_MASS_EV
    elif expected_mass is not None and _mass_consistent(pdg_id, expected_mass):
        # This is a workaround in case an exact mass is given
        # (like for the reference particle)
        return expected_mass
    else:
        raise ValueError(f"Exact mass not found for particle with PDG ID {pdg_id}.")


# Dynamically set _pdg_inv so it can be defined only here
_PDG_inv = {}
def get_pdg_inv():
    if _PDG_inv == {}:
        for pdg_id, val in pdg_table.items():
            for vv in val[1:]:
                _PDG_inv[_to_normal_script(vv).lower()] = pdg_id
    return _PDG_inv


def _flip_end_sign(name):
    if name[-2:] == '⁺⁺':
        return name[:-2] + '⁻⁻'
    elif name[-2:] == '++':
        return name[:-2] + '--'
    elif name[-2:] == '⁻⁻':
        return name[:-2] + '⁺⁺'
    elif name[-2:] == '--':
        return name[:-2] + '++'
    elif name[-1] == '⁺':
        return name[:-1] + '⁻'
    elif name[-1] == '+':
        return name[:-1] + '-'
    elif name[-1] == '⁻':
        return name[:-1] + '⁺'
    elif name[-1] == '-':
        return name[:-1] + '+'
    else:
        return name


def _digits_to_superscript(val):
    val = f'{val}'.replace('0', '⁰').replace('1', '¹').replace('2', '²').replace('3', '³')
    val = val.replace('4', '⁴').replace('5', '⁵').replace('6', '⁶').replace('7', '⁷')
    return val.replace('8', '⁸').replace('9', '⁹').replace('+', '⁺').replace('-', '⁻')

def _digits_to_normalscript(val):
    val = f'{val}'.replace('⁰', '0').replace('¹', '1').replace('²', '2').replace('³', '3')
    val = val.replace('⁴', '4').replace('⁵', '5').replace('⁶', '6').replace('⁷', '7')
    return val.replace('⁸', '8').replace('⁹', '9').replace('⁺', '+').replace('⁻', '-')

def _to_normal_script(val):
    return _digits_to_normalscript(val).replace('ₛ', 's').replace('ₑ', 'e')


# Dynamically set mass_table from xtrack.particles.masses to avoid circular imports and loops
_mass_table = {}
def _get_mass_table():
    if _mass_table == {}:
        for name, mass in mass__dict__.items():
            if name.endswith('_MASS_EV') and name != 'U_MASS_EV':
                # Mass definitions for charged particles are without the charge identifier,
                # so we need to try these first.
                try:
                    pdg_id = get_pdg_id_from_name(f'{name[:-8]}+')
                except ValueError:
                    pdg_id = get_pdg_id_from_name(name[:-8])
                if pdg_id not in _mass_table:
                    _mass_table[pdg_id] = mass
                else:
                    raise ValueError(f"Duplicate mass in Xtrack for particle {pdg_id}!")
    return _mass_table


def _mass_consistent(pdg_id, m, mask=None):
    if hasattr(pdg_id, '__len__') and not isinstance(pdg_id, str) \
    and hasattr(m, '__len__') and not isinstance(m, str):
        if mask is None:
            return np.all([_mass_consistent(pdg, mm) for pdg, mm in zip(pdg_id, m)])
        else:
            pdg_id = np.array(pdg_id)[mask]
            m = np.array(m)[mask]
            return np.all([_mass_consistent(pdg, mm) for pdg, mm in zip(pdg_id, m)])
    elif hasattr(pdg_id, '__len__') and not isinstance(pdg_id, str):
        if mask is None:
            return np.all([_mass_consistent(pdg, m) for pdg in pdg_id])
        else:
            pdg_id = np.array(pdg_id)[mask]
            return np.all([_mass_consistent(pdg, m) for pdg in pdg_id])
    elif hasattr(m, '__len__') and not isinstance(m, str):
        if mask is None:
            return np.all([_mass_consistent(pdg_id, mm) for mm in m])
        else:
            m = np.array(m)[mask]
            return np.all([_mass_consistent(pdg_id, mm) for mm in m])

    try:
        q, _, _, _ = get_properties_from_pdg_id(pdg_id)
        return pdg_id == get_pdg_id_from_mass_charge(m, q)
    except ValueError:
        # No check if mass cannot be retrieved
        return True


# Make sure no duouble names exist in the pdg_table, after removing subscripts
# and going to lower case
def _check_pdg_table():
    names = [vvv for vv in pdg_table.values() for vvv in vv[1:]]
    names = [_to_normal_script(name).lower() for name in names]
    if len(names) != len(np.unique(names)):
        double_names = []
        for i, name1 in enumerate(names):
            for name2 in names[i+1:]:
                if _to_normal_script(name1).lower() \
                == _to_normal_script(name2).lower():
                    double_names.append(name1)
                    double_names.append(name2)
        raise ValueError(f"Duplicate names in pdg_table:\n{double_names}")

_check_pdg_table()
